/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.agent;

import java.util.LinkedList;
import java.util.TreeMap;

/**
 * @author mike
 * This will be the interface between programs that use the framework and the actual lojban processing.
 * This level will store items such as prosumti assignments that are local to the current speaker,
 * but it will delegate to the knowledge base for any factual knowledge.
 */
public class LojbanContext {
	
	private TreeMap<String, LojbanPlugin> plugins = new TreeMap<String, LojbanPlugin>();
	private TreeMap<String, Boolean> pluginState = new TreeMap<String, Boolean>();
	
	private LinkedList<IConversationListener> convers = new LinkedList<IConversationListener>();
	
	private LojbanKnowledgeBase lkb = null;
	private boolean multiple = false;
	
	public LojbanContext(LojbanKnowledgeBase lkb) {
		this.lkb = lkb;
	}
	
	public String[] listPlugins() {
		return plugins.keySet().toArray(new String[]{});
	}
	
	public boolean getPluginStatus(String s) {
		return pluginState.get(s);
	}
	
	public void addPlugin(LojbanPlugin p) {
		plugins.put(p.getPluginName(), p);
		pluginState.put(p.getPluginName(), false);
	}
	
	public void setPluginStatus(String s, boolean b) {
		boolean old = pluginState.get(s);
		if(old != b) {
			pluginState.put(s, b);
			if(b)
				plugins.get(s).load(this);
			else
				plugins.get(s).unload();
		}
	}
	
	/*public void addRelationListener(String relation, IRelationListener rl) {
		
	}
	
	public void addSubjectListener(String subject, ISubjectListener sl) {
		
	}*/
	
	/**
	 * Add a listener to receive conversation notices.
	 * @param cl the listener object
	 */
	public void addConversationListener(IConversationListener cl) {
		convers.add(cl);
	}
	
	/**
	 * Remove conversation listener.
	 * @param cl
	 */
	public void removeConversationListener(IConversationListener cl) {
		convers.remove(cl);
	}
	
	/**
	 * Send a line of lojban text through the engine
	 * @param sIn input string
	 * @return output/response string
	 */
	public String speak(String sIn) {
		
		sendStatus("Running input plugins...");
		
		// send input to plugins
		for(IConversationListener cl : convers)
			cl.input(sIn);
		
		String sOut = null;
		try {
			// send to knowedgebase
			sendStatus("Processing query...");
			lkb.speak(sIn, this);
			
		} catch(Exception e) {
			// send error to plugins
			sendStatus("Error!");
			for(IConversationListener cl : convers)
				cl.error(e.getMessage());
			return null;
		}
						
		sendStatus("Ready.");
		
		return sOut;
	}
	
	public void sendStatus(String s) {
		// send status to plugins
		for(IConversationListener cl : convers)
			cl.status(s);
	}
	
	public void sendOutput(String s) {
		sendStatus("Running output plugins...");
		// send status to plugins
		for(IConversationListener cl : convers)
			cl.output(s);
	}

	public boolean isMultiple() {
		return multiple;
	}
	
	public void setMultiple(boolean m) {
		multiple = m;
	}
	
	
	/*
	public void tellFact(IFact f) {
		
	}
	
	public ISubject[] askFact(IFact f) {
		
	}
	
	public IFact[] describeSubject(ISubject s) {
		
	}
	*/
}
