/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.agent;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import lojban.analyzer.Analysis;
import lojban.analyzer.AnalysisException;
import lojban.analyzer.IFact;
import lojban.analyzer.IRelation;
import lojban.analyzer.ISubject;
import lojban.kb.jtp.JTPFact;
import lojban.kb.jtp.JTPResolver;
import lojban.kb.jtp.JTPSubject;
import lojban.parser.ParseException;
import lojban.synthesis.Synthesis;

/**
 * @author mike
 * Contains the actual knowledgebase itself.  Contexts delegate here for any real information processing.
 */
public class LojbanKnowledgeBase {

	private Analysis a = new Analysis();
	private Synthesis s = new Synthesis();
	private JTPResolver resolver = new JTPResolver();
		
	private PrintStream errorLog = null;
	
	/**
	 * Create knowledge base and open error log.
	 */
	public LojbanKnowledgeBase() {
		try {
			errorLog = new PrintStream(new FileOutputStream(new File("error.log")));
		} catch(Exception e) {
			System.err.println("couldn't open error log");
		}
	}
	
	/**
	 * Speak a string to the knowledgebase
	 * @param in input string
	 * @return any reply, or null if no reply (input was a statement/command which didn't demand one)
	 */
	public void speak(String in, LojbanContext lc) {
		
		try {
				
			// analyze and resolve to facts
			lc.sendStatus("Analyzing...");
			a.analyze(in);
			JTPFact fact = (JTPFact)a.resolve(resolver);
						
			if(fact.isQuery() && !fact.isConfirmQuery()) {			// HANDLE FILL IN THE BLANKS QUERIES
					
				lc.sendStatus("Running info query...");
								
				// process query
				Set<Map<ISubject,ISubject[]>> answers=resolver.ask(fact, lc.isMultiple());
				
				boolean noAnswers = true;
				
				for(Map<ISubject,ISubject[]> answer : answers){
					
					// reform the question with the answers so the describe algorithm
					// knows not to repeat parts of it
					IFact question = fact;
					for(ISubject key : answer.keySet()) {
						question = question.substitute(key, answer.get(key)[0]);
					}
					
					// get descriptions for all facts
					HashMap<ISubject, ISubject> descriptions = new HashMap<ISubject, ISubject>();
					HashMap<IRelation, IRelation> selbri = new HashMap<IRelation, IRelation>();
					lc.sendStatus("Describing results...");					
					for(ISubject key : answer.keySet())
						
						if(key.toString().startsWith("?askrel")) {		// answer to a mo query
							
							selbri.put(resolver.getRelationByName(key.toString()),
									   resolver.getRelationByName(answer.get(key)[0].toString()));
							
						} else if(key.toString().startsWith("?ask")) {	// answer to a ma query
							
							ISubject subject = answer.get(key)[0];
							ISubject described = subject;
							
							if(subject.toString().startsWith("zohe"))
								described = resolver.getProsumtiSubject("zo'e");
							else if(!subject.isLiteral())
								described = resolver.describe((JTPSubject)subject, question);
							
							if(key != null && described != null)
								descriptions.put(key, described);
							else {
								System.out.println("null description?");
							}
						}
					lc.sendStatus("Synthesizing response...");
					lc.sendOutput(s.synthesizeFact(s.createAnswerFact(fact, descriptions, selbri, resolver), resolver));
					if(!lc.isMultiple())
						return;
					else
						noAnswers = false;
				}
				
				if(noAnswers)
					lc.sendOutput("na go'i");
				
			} else if(fact.isConfirmQuery() && !fact.isQuery()) {		// HANDLE TRUE/FALSE QUERIES
	
				lc.sendStatus("Running true/false query...");
				
				// process query
				Set<Map<ISubject,ISubject[]>> answers=resolver.ask(fact, lc.isMultiple());
				if(answers.size() > 0)
					lc.sendOutput("go'i");
				else
					lc.sendOutput("na go'i");
				
			} else if(fact.isConfirmQuery() && fact.isQuery()) {
				
				throw new KnowledgeBaseException("cannot resolve mo/xu queries simultaneously");
				
			} else {
					lc.sendStatus("Telling knowledge base...");
					resolver.tell(fact);
			}
		
		} catch(AnalysisException e) {
			e.printStackTrace(errorLog);
			throw new KnowledgeBaseException("error during analysis: " + e.getMessage());
		} catch(ParseException e) {
			e.printStackTrace(errorLog);
			throw new KnowledgeBaseException("error during parsing: " + e.getMessage());
		} catch(KnowledgeBaseException e) {
			e.printStackTrace(errorLog);
			throw new KnowledgeBaseException("knowledgebase error: " + e.getMessage());
		} catch(Exception e) {
			e.printStackTrace(errorLog);
			throw new KnowledgeBaseException(e.getClass().getName() + ": " + e.getMessage());
		}
	}

	/**
	 * @return a new context to this knowledgebase
	 */
	public LojbanContext getContext() {
		return new LojbanContext(this);
	}
}
