/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.analyzer;

import lojban.parser.GenericParseNode;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;

abstract class AbstractRecursiveAnalysis {
	
	private String text = null;
	
	public void process(GenericParseNode n) {
		
		// save name if this is the first entry to the analyzer
		boolean start = false;
		if(text == null) {
			text = n.getText();
			start = true;
			start();				// call starting function
		}
		
		String s = getFunctionName(n.getType());
		
		try {
			Method m = getClass().getMethod(s, new Class[]{GenericParseNode.class});
			m.invoke(this, new Object[]{n});
		} catch (NoSuchMethodException e) {
			throw new AnalysisException("no method to handle " + s + " on " + getDebugName());
		} catch (SecurityException e) {
			e.printStackTrace();
			throw new AnalysisException("unknown analysis problem");
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
			throw new AnalysisException("unknown analysis problem");
		} catch (IllegalAccessException e) {
			e.printStackTrace();
			throw new AnalysisException("unknown analysis problem");
		} catch (InvocationTargetException e) {
			if(e.getCause() instanceof AnalysisException)
				throw (AnalysisException)e.getCause();
			e.printStackTrace();
			throw new AnalysisException("unknown analysis problem: " + e.getCause());
		}
		
		if(start)
			end();					// call ending function
	}
	
	/**
	 * Called before processing begins.  Override for extra setup.
	 */
	public void start() {
		
	}
	
	/**
	 * Called as processing ends.  Override for extra cleanup.
	 */
	public void end() {
		
	}
	
	public void process(Collection<GenericParseNode> nc) {
		for(GenericParseNode n1 : nc)
			process(n1);
	}
	
	public void process(GenericParseNode n[]) {
		if(n != null)
			for(GenericParseNode n1 : n)
				if(n1 != null)
						process(n1);
	}
	
	public void processChildren(GenericParseNode n) {
		process(n.getChildren());
	}
	
	/**
	 * Get text this analyzer is responsible for. 
	 */
	public String getText() {
		return text;
	}
	
	/**
	 * Get a friendlier name for a funciton.
	 * @param s the type name
	 * @return a function name
	 */
	private String getFunctionName(String s) {
		return "process" + Character.toUpperCase(s.charAt(0)) + s.toLowerCase().substring(1);
	}
	
	/**
	 * String to refer to this class as in debug messages
	 * @return debug string
	 */
	public abstract String getDebugName();
	
}
