/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.analyzer;

import java.util.Vector;

import lojban.parser.AbstractDictionary;
import lojban.parser.GenericParseNode;

/**
 * @author mike
 * Interface class to break lojban text down into facts, relations, and subjects.
 * First provide a string with analyze() to build a semantic structure internally.
 * Then the other functions may be used to examine the result.
 * Additional strings may then be provided for analysis within the same context
 * (i.e., pronoun assignemnts) as the previous strings.
 */
public class Analysis {
	
	private TextAnalysis ta = null;
	private String text = null;
	Vector<GenericParseNode> v = null;
	
	public Analysis() {
		
	}
	
	/**
	 * @param s
	 */
	public Analysis(String s) {
		analyze(s);
	}

	public void analyze(String s) {
		
		text = s;
			
		// create new analysis
		ta = new TextAnalysis(ta);
		
		// parse nodes
		v = GenericParseNode.parse(s);

		// run analysis on nodes
		ta.process(v);
	}
	
	public IFact resolve(IResolver resolver) {
		
		if(ta == null)
			throw new AnalysisException("no string analysis data");
		
		IFact f[] = ta.resolve(resolver);
		IFact result = null;
		for(IFact f1 : f) {
			if(result == null)
				result = f1;
			else
				result = resolver.createConnectiveFact(IFact.Connective.AND, false, false, result, f1);
		}
		return result;
		
	}
	
	public String toTranslation(AbstractDictionary ad) {
		
		if(ta == null)
			throw new AnalysisException("no string analysis data");
		
		return ta.toTranslation(ad);
	}
	
	public String toString() {
		
		if(ta == null)
			throw new AnalysisException("no string analysis data");
		
		return ta.toString();
	}
	
	public String getParseTree() {
		
		if(ta == null)
			throw new AnalysisException("no string analysis data");
		
		// debug output
		StringBuilder b = new StringBuilder();
		boolean start = true;
		for(GenericParseNode n : v) {
			if(!start)
				b.append("\r\n");
			start = false;
			b.append(n.toString());
		}
		return b.toString().trim();
	}

	public String getText() {
		
		if(ta == null)
			throw new AnalysisException("no string analysis data");
		
		return text;
	}
}
