/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.analyzer;

import lojban.agent.KnowledgeBaseException;

public interface IResolver {
	
	/**
	 * Create a new fact object
	 * @param isTrue true if the fact is true
	 * @param r relation (selbri)
	 * @param s1 subject in x1 place (sumti)
	 * @param s2 subject in x2 place (sumti)
	 * @param s3 subject in x3 place (sumti)
	 * @param s4 subject in x4 place (sumti)
	 * @param s5 subject in x5 place (sumti)
	 * @return fact as described by arguments.
	 */
	public IFact createFact(boolean isTrue, IRelation r, ISubject s1, ISubject s2, ISubject s3, ISubject s4, ISubject s5);
	
	/**
	 * Create a fact consisting of a logical connective between two facts. 
	 * @param c connective type
	 * @param fact1 left hand fact
	 * @param fact2 right hand fact
	 * @param neg1 true if first fact is negated
	 * @param neg2 true if second fact is negated
	 * @return
	 */
	public IFact createConnectiveFact(IFact.Connective c, boolean neg1, boolean neg2, IFact fact1, IFact fact2);
	
	/**
	 * Find relation by string.  If none found, instantiate one and tag it with the right string.
	 * @param name name to search for
	 * @return a matching relation
	 */
	public IRelation getRelationByName(String name);
	
	/**
	 * Find subject by name (cmene).  If none found, instantiate one and tag it with the right name.
	 * @param name name to search for
	 * @param querySubject 
	 * @return a matching subject
	 */
	public ISubject getSubjectByName(String name, IFact f[], ISubject querySubject);
			
	/**
	 * Find a subject given restricting facts.  Note that some of these restricting facts might be "is associated by"
	 * (determined by relation) and will need special handling ultimately.
	 * @param r relation the result should be in the x1 spot for
	 * @param f facts with query objects that a match must fulfill
	 * @param allowNew if true, this function must instantiate a matching subject if none is found, otherwise it must return null if none is found
	 * @return a matching subject
	 */
	public ISubject getSubjectByRestrictions(IRelation r, IFact f[], ISubject q, boolean veridical);

	/**
	 * @return the relative clause placeholder subject
	 */
	public ISubject getRelativeSubject();
	
	/**
	 * @return the query placeholder subject
	 */
	public ISubject getQuerySubject();
	
	/**
	 * Get a subject for a string literal
	 * @param literal
	 * @return subject for literal
	 */
	public ISubject getLiteralSubject(String literal);
	
	/**
	 * Push this assignment onto the prosumti stack.
	 * @param prosumti the prosumti
	 * @param s the new subject assignment
	 */
	public void pushAssignment(String prosumti, ISubject s);
	
	/**
	 * Remove the current assignment for a prosumti.
	 * @param prosumti the prosumti
	 */
	public void popAssignment(String prosumti);
	
	/**
	 * @param prosumti
	 * @return the current assignemnt for the pronound (throw exception if unassigned)
	 */
	public ISubject getAssignment(String prosumti);

	/**
	 * Get a subject that represents a quote.
	 * @param fact the facts of the quote
	 * @return a subject that refers to the quote
	 */
	public ISubject getQuoteSubject(IFact fact);
	
	/**Describe a subject with facts
	 * @param subject the subject to describe
	 * @return a fact that uniquely describes the subject (does not describe any other subjects)
	 * @throws KnowledgeBaseException
	 */
	public ISubject describe(ISubject subject) throws KnowledgeBaseException;

	/**
	 * @return query object for a relation
	 */
	public IRelation getQueryRelation();

	/**
	 * @return a fact which is a query to confirm its statement
	 */
	public IFact createConfirmFact(boolean b, IRelation r, ISubject subject, ISubject subject2, ISubject subject3, ISubject subject4, ISubject subject5);
}
