/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.analyzer;

import java.util.Vector;

import lojban.parser.AbstractDictionary;
import lojban.parser.GenericParseNode;

/**
 * @author mike
 * Analyzes the nodes which indicate a Sumti to resolve it to an object.
 */
class ReferenceAnalysis extends AbstractRecursiveAnalysis {
	
	/** type of reference */
	private enum ReferenceType {CMENE, VERIDICAL, NONVERIDICAL, QUERY, PROSUMTI, QUOTE, LITERAL, EXPRESSION};
	private ReferenceType type = null;
	private String prosumti = null;
	
	/** cmene (name) information */
	private String cmene = null;
	
	/** selbri (verb) information */
	private RelationAnalysis relation = null;
	
	/** relative clauses which restrict reference */
	private Vector<RelativeClauseAnalysis> restrictives = new Vector<RelativeClauseAnalysis>();
	
	/** relative clauses which are incidental to reference */
	private Vector<RelativeClauseAnalysis> incidentals = new Vector<RelativeClauseAnalysis>();
	
	/** in case of prosumti, this is antecedent */
	private ReferenceAnalysis antecedent = null;
	
	/** whether this sumti can be the antecedent of an anaphoric prosumti */
	private boolean isAnaphoricAntecedent = false;
	
	/** whether this sumti is complete (and qualifies for anaphoric reference */
	private boolean complete = false;
	
	/** if this is >=0, the target of a reflexive prosumti */
	private int reflexive = -1;
	
	/** if this is a grammatical quote */
	private Analysis quoteAnalysis = null;
	
	/** if this is a literal quote */
	private String quoteLiteral = null;
	
	/** if this is a mathematical expression */
	@SuppressWarnings("unused")
	private MathAnalysis expression = null;
	
	/** cached resolution */
	private ISubject resolution = null;
	
	private TextAnalysis ta = null;
	
	public ReferenceAnalysis(TextAnalysis ta) {
		this.ta = ta;
	}
	
	/**
	 * Get a resolution for this reference.
	 * @param resolver resolver object
	 * @return a subject
	 */
	public ISubject resolve(IResolver resolver, Vector<IFact> facts) {
		
		// if already resolved, return the resolution
		if(resolution != null)
			return resolution;
		
		// if we have an antecedent, delegate to it
		if(antecedent != null)
			return antecedent.resolve(resolver, facts);
		
		ISubject subject = null;
		
		Vector<IFact> restrictingFacts = null;
		
		// now find out who/what/where/etc we reference
		switch(type) {
		
		case EXPRESSION:
			// TODO: handling of mekso (not implemented) could be attached here
			subject = resolver.getLiteralSubject("expression");
			break;
		
		case LITERAL:
			subject = resolver.getLiteralSubject(quoteLiteral);
			break;
		
		case QUOTE:
			subject = resolver.getQuoteSubject(quoteAnalysis.resolve(resolver));
			break;
		
		case PROSUMTI:
			// this will be prosumti without antecedents (antecedents already delegated above)
			subject = resolver.getAssignment(prosumti);
			break;
		
		case CMENE:					// NAME
			
			// resolve all restricting clauses
			restrictingFacts = new Vector<IFact>();
			ISubject querySubject = resolver.getRelativeSubject();
			for(RelativeClauseAnalysis rca : restrictives)
				rca.resolveRestrictive(resolver, restrictingFacts, facts, querySubject);
			
			String noun = cmene;
			if(noun == null)
				noun = relation.getText();
			
			// get subject
			subject = resolver.getSubjectByName(noun, restrictingFacts.toArray(new IFact[]{}), querySubject);
			break;
		
		case QUERY:					// QUERY
			subject = resolver.getQuerySubject();
			break;
			
		case VERIDICAL:				// LE or LO
		case NONVERIDICAL:
			
			// get relation
			IRelation r = relation.resolve(resolver);
						
			// get query subject placeholder
			ISubject querySubj = resolver.getRelativeSubject();
			
			// resolve all restricting clauses
			restrictingFacts = new Vector<IFact>();
			for(RelativeClauseAnalysis rca : restrictives)
				rca.resolveRestrictive(resolver, restrictingFacts, facts, querySubj);
			
			// try to get a subject
			subject = resolver.getSubjectByRestrictions(
					r,
					restrictingFacts.toArray(new IFact[]{}),
					querySubj,
					type == ReferenceType.VERIDICAL);
			
			break;			
			
		default:
			throw new AnalysisException("unable to resolve (unknown type?): " + getText());
		
		}
		
		// ensure subject's not null
		if(subject == null)
			throw new AnalysisException("unable to resolve (no matches?): " + getText());

		// add incidental clauses
		for(RelativeClauseAnalysis rca : incidentals)
			rca.resolveIncidental(resolver, subject, facts);
		
		resolution = subject;
		
		return subject;
		
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.AbstractRecursiveAnalysis#getDebugName()
	 */
	public String getDebugName() {
		return "Reference";
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.AbstractRecursiveAnalysis#start()
	 */
	public void start() {
		ta.recordSumti(this);		// register for anaphoric reference
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.AbstractRecursiveAnalysis#end()
	 */
	public void end() {
		complete = true;			// mark done (thus eligible for reference)
	}
	
	public void processSumti(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processSumti1(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processSumti2(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processSumti3(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processSumti4(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processSumti5(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processSumti6(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Handle prosumti by finding antecedent.
	 * @param n
	 */
	public void processKohaclause(GenericParseNode n) {
		type = ReferenceType.PROSUMTI;
		prosumti = n.getText();
		
		// assignable prosumti
		if(prosumti.startsWith("ko'") || prosumti.startsWith("fo'")) {
			antecedent = ta.getAssignedRef(prosumti);
			isAnaphoricAntecedent = false;
		}
		
		// "last" prosumti
		if(prosumti.equals("ri")) {
			antecedent = ta.getAnaphoricRef();
			isAnaphoricAntecedent = true;
		}
		
		// ambiguous anaphorics
		if(prosumti.equals("ra") || prosumti.equals("ru"))
			throw new AnalysisException("ra/ru cannot be supported due to ambiguity");
		
		// reflexives
		if(prosumti.startsWith("vo'")) {
			reflexive = "aeiou".indexOf(prosumti.charAt(3));
		}
		
		// query
		if(prosumti.equals("ma")) {
			type = ReferenceType.QUERY;
		}
		
		// ke'a/personals go through this?
	}
	
	/**
	 * Record seeing a la clause, this is cmene
	 * @param n
	 */
	public void processLaclause(GenericParseNode n) {
		isAnaphoricAntecedent = true;
		type = ReferenceType.CMENE;
	}
	
	/**
	 * Record seeing a le/lo clause, this is (non)veridical
	 * @param n
	 */
	public void processLeclause(GenericParseNode n) {
		isAnaphoricAntecedent = true;
		if(n.getText().equals("le"))
			type = ReferenceType.NONVERIDICAL;
		else if(n.getText().equals("lo"))
			type = ReferenceType.VERIDICAL;
		else
			throw new AnalysisException("unexpected le clause: " + n.getText());
	}
	
	/**
	 * This is a quotation
	 * @param n
	 */
	public void processLuclause(GenericParseNode n) {
		type = ReferenceType.QUOTE;
	}
	
	/**
	 * Single world literal quote
	 * @param n
	 */
	public void processZoclause(GenericParseNode n) {
		type = ReferenceType.LITERAL;
		quoteLiteral = n.getChildren()[0].getChildren()[1].getText().trim();
	}
	
	/**
	 * A quotation
	 * @param ParseNode
	 */
	public void processText(GenericParseNode n) {
		quoteAnalysis = new Analysis(n.getText());
	}
	
	/**
	 * Quote terminator, safe to ingore.
	 * @param n
	 */
	public void processLihuclause(GenericParseNode n) {
		
	}
	
	/**
	 * Lojban literal multiword quote.
	 * @param n
	 */
	public void processLohuclause(GenericParseNode n) {
		type = ReferenceType.LITERAL;
		GenericParseNode n1[] = n.getChildren()[0].getChildren();
		for(int i = 1; i < n1.length-1; i++) {
			if(quoteLiteral == null)
				quoteLiteral = n1[i].getText().trim();
			else
				quoteLiteral += " " + n1[i].getText().trim();
		}
	}
	
	/**
	 * Non-lojban quote.
	 * @param n
	 */
	public void processZoiclause(GenericParseNode n) {
		type = ReferenceType.LITERAL;
		GenericParseNode n1[] = n.getChildren()[0].getChildren();
		for(int i = 2; i < n1.length-1; i++) {
			if(quoteLiteral == null)
				quoteLiteral = n1[i].getText().trim();
			else
				quoteLiteral += " " + n1[i].getText().trim();
		}
	}
	
	public void processCmeneclause(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processCmenepre(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Record cmene information
	 * @param n
	 */
	public void processCmene(GenericParseNode n) {
				
		// if "lowercase" cmene, save the word
		if(n.getType().equals("cmene")) {
			if(cmene == null)
				cmene = n.getContent();
			else
				cmene += " " + n.getContent();
		} else // recurse
			processChildren(n);
	}
	
	/**
	 * There could be a posessive in the form "le [la maik.] karce" here.
	 * If so, strip it out as a restrictive clause.
	 * @param n
	 */
	public void processSumtitail(GenericParseNode n) {
		
		for(GenericParseNode n1 : n.getChildren()) {
			
			if(n1.getType().equals("sumtiTail1"))	// the actual sumti here
				process(n1);
			
			else {									// a possessive node
				if(n1.getType().equals("sumti6")) {
					
					// delegate to analysis as a restrictive clause
					RelativeClauseAnalysis rca = new RelativeClauseAnalysis(ta);
					rca.process(n1);
					restrictives.add(rca);
				} else {
					throw new AnalysisException("unexpected " + n1.getType() + ": " + n1.getText());
				}
			}
		}
	}
	
	public void processSumtitail1(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Delegate selbri to RelationAnalysis
	 * @param n
	 */
	public void processSelbri(GenericParseNode n) {
		
		// delegate construction of the selbri (probably in the sumti tail)
		relation = new RelationAnalysis(ta);
		relation.process(n);
	}
	
	public void processRelativeclauses(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Delegate relative clauses to RelativeClauseAnalysis
	 * @param n
	 */
	public void processRelativeclause(GenericParseNode n) {
		
		// delegate to RelativeClauseAnalysis
		RelativeClauseAnalysis rca = new RelativeClauseAnalysis(ta);
		rca.process(n);
		
		if(rca.isAssignment()) {
			ta.makeAssignment(rca.getProsumti(), this);
		} else if(rca.isRestrictive()) {
			restrictives.add(rca);
		} else
			incidentals.add(rca);
		
	}
	
	/**
	 * May be safely ignored.  Only needed by parser.
	 * @param n
	 */
	public void processZiheclause(GenericParseNode n) {
		
	}
	
	/**
	 * Delegate handling of mathematical expressions.
	 * @param n
	 */
	public void processLiclause(GenericParseNode n) {
		type = ReferenceType.EXPRESSION;
		expression = new MathAnalysis();
		//expression.parseExpression(n.getChildren()[1]);
	}
	
	/**
	 * Ignore ku clause--needed for parser only
	 * @param n
	 */
	public void processKuclause(GenericParseNode n) {
		// safe to ignore
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		
		// delegate to antecedent first
		if(antecedent != null)
			return "{" + antecedent.toString() + "}";
		
		StringBuilder sb = new StringBuilder();
		
		switch(type) {
		
		case EXPRESSION:
			sb.append("#EXPRESSION#");
			break;
		
		case LITERAL:
			sb.append("'" + quoteLiteral + "'");
			break;
		
		case QUOTE:
			sb.append("<" + quoteAnalysis.toString() + ">");
			break;
		
		case PROSUMTI:
			// this is specifically prosumti without antecedents (or they'd have been handled)
			sb.append(prosumti);
			break;
		
		case QUERY:
			sb.append("?");
			break;
			
		case CMENE:
			if(cmene != null)
				sb.append(cmene);
			else
				sb.append(relation.toString());
			break;
			
		case VERIDICAL:
		case NONVERIDICAL:
			
			sb.append(relation);			
			break;
		
		default:
			sb.append("#ERROR:RefA#");
		
		}
		
		// add relative clauses
		sb.append("[");
		for(RelativeClauseAnalysis rca : restrictives)
			sb.append(rca + ", ");
		if(restrictives.size() > 0)
			sb.delete(sb.length()-2, sb.length());	// remove extra commas
		sb.append("|");
		for(RelativeClauseAnalysis rca : incidentals)
			sb.append(rca + ", ");
		if(incidentals.size() > 0)
			sb.delete(sb.length()-2, sb.length());	// remove extra commas
		sb.append("]");
		
		return sb.toString();
	}

	/**
	 * @return approximation in English
	 */
	public String toTranslation(AbstractDictionary ad) {
		
		// delegate to antecedent first
		if(antecedent != null)
			return antecedent.toTranslation(ad);
		
		StringBuilder sb = new StringBuilder();
		
		switch(type) {
		
		case EXPRESSION:
			sb.append("#EXPRESSION#");
			break;
		
		case LITERAL:
			sb.append("'" + quoteLiteral + "'");
			break;
		
		case QUOTE:
			sb.append("<" + quoteAnalysis.toTranslation(ad) + ">");
			break;
		
		case PROSUMTI:
			// this is specifically prosumti without antecedents (or they'd have been handled)
			sb.append(ad.getWord(prosumti));
			break;
		
		case QUERY:
			sb.append(ad.getWord("ma"));
			break;
			
		case CMENE:

			String noun = cmene;
			if(noun == null)
				noun = ad.getNoun(relation.toString());

			sb.append(Character.toUpperCase(noun.charAt(0)));
			sb.append(noun.substring(1) + " ");
			break;

		case VERIDICAL:
		case NONVERIDICAL:
			
			// get article
			if(type == ReferenceType.VERIDICAL)
				sb.append(ad.getWord("la") + " ");
			else
				sb.append(ad.getWord("le") + " ");
			
			// render relation word
			sb.append(ad.getNoun(relation.toString()) + " ");
			
			break;
		
		default:
			sb.append("#ERROR:RefA#");
		
		}
		
		// render relative clauses
		if(restrictives.size() + incidentals.size() > 0) {
			sb.append("(");
				
			// render extra clauses
			for(RelativeClauseAnalysis rca : restrictives)
				sb.append(rca.toTranslation(ad) + "; ");
			for(RelativeClauseAnalysis rca : incidentals)
				sb.append(rca.toTranslation(ad) + "; ");
			sb.delete(sb.length()-2, sb.length());	// remove extra commas
			
			sb.append(")");
		}
		
		return sb.toString();
	}

	/**
	 * @return true if this sumti is complete
	 */
	public boolean isComplete() {
		return complete;
	}

	/**
	 * @return true if this sumti is legal for anaphoric reference
	 */
	public boolean isAnaphoricAntecedent() {
		return isAnaphoricAntecedent;
	}
	
	/**
	 * @return -1 if not reflexive prosumti, else 0-4 for target sumti
	 */
	public int getReflexive() {
		return reflexive;
	}

	/**
	 * Set an antecedent for this item (used for reflexives)
	 * @param reflexive
	 */
	public void setAntecedent(ReferenceAnalysis r) {
		antecedent = r;
	}

	/**
	 * Get the digit associated with a prosumti.
	 * @param prosumti2
	 * @return digit associated with prosumti
	 */
	public static String getProsumtiNumber(String prosumti) {
		int minor = "aeiou".indexOf(prosumti.charAt(3)) + 1;
		if(prosumti.startsWith("ko"))
			return minor + "";
		else
			return (5+minor) + "";
	}

	/**
	 * @return true if this is a ke'a--a pointer in a relative clause
	 */
	public boolean isKeha() {
		return prosumti != null && prosumti.equals("ke'a");
	}
}
