/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.analyzer;

import java.util.Vector;

import lojban.parser.GenericParseNode;
import lojban.parser.AbstractDictionary;

class RelativeClauseAnalysis extends AbstractRecursiveAnalysis {

	/** type of relative clause */
	private enum RelativeType { PE, PO, NOI, POI, GOI };
	private RelativeType type;
	
	/** assignable prosumti if this is a goi clause */
	private String prosumti;
	
	private ReferenceAnalysis subject = null;
	private SentenceAnalysis sentence = null;
	
	private TextAnalysis ta = null;
	
	public RelativeClauseAnalysis(TextAnalysis ta) {
		this.ta = ta;
	}
	
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.AbstractRecursiveAnalysis#getDebugName()
	 */
	public String getDebugName() {
		return "RelativeClause";
	}
	
	/**
	 * This function will be the entry point (rather than processRelativeclause) if this is an implicit posessive
	 * (which reference analysis would delegate to here to get a restrictive clause)
	 * @param n
	 */
	public void processSumti6(GenericParseNode n) {
		
		// delegate to reference analysis to figure out the owner
		subject = new ReferenceAnalysis(ta);
		subject.process(n);
		
		// call this a PE
		type = RelativeType.PE;
	}
	
	public void processRelativeclause(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processRelativeclause1(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processTerm(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processTerm1(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Delegate a subject (for a possessive/associative (GOI) relationship) to a ReferenceAnalysis.
	 * @param n
	 */
	public void processSumti(GenericParseNode n) {
		
		if(type == RelativeType.GOI)		// get prosumti for assignment
			prosumti = n.getText();
		else {								// process as normal sumti
			assert subject == null;
			subject = new ReferenceAnalysis(ta);
			subject.process(n);
		}	
	}
	
	/**
	 * Handle possession/association.
	 * @param n
	 */
	public void processGoiclause(GenericParseNode n) {
		String s = n.getText();
		
		if(s.equals("pe"))
			type = RelativeType.PE;
		else if(s.equals("po"))
			type = RelativeType.PO;
		else if(s.equals("goi")) {
			type = RelativeType.GOI;
		} else
			throw new AnalysisException("unrecognized GOI \'" + s + "\' in relative clause: " + getText()); 
	}
	
	/**
	 * Handle possession/association.
	 * @param n
	 */
	public void processNoiclause(GenericParseNode n) {
		String s = n.getText();
		
		if(s.equals("noi"))
			type = RelativeType.NOI;
		else if(s.equals("poi"))
			type = RelativeType.POI;
		else
			throw new AnalysisException("unrecognized NOI \'" + s + "\' in relative clause: " + getText()); 
	}
	
	/**
	 * Handle a sentence on a noi/poi clause by delegating to SentenceAnalysis.
	 * @param n
	 */
	public void processSentence(GenericParseNode n) {
		assert sentence == null;
		sentence = new SentenceAnalysis(ta, false, false);
		sentence.process(n);
	}
	
	public void processSubsentence(GenericParseNode n) {
		processChildren(n);
	}


	/**
	 * @return true if this clause is of a restrictive type
	 */
	public boolean isRestrictive() {
		return
			type == RelativeType.POI ||
			type == RelativeType.PE ||
			type == RelativeType.PO;
	}
	
	/**
	 * @return true if this is a prosumti assignment
	 */
	public boolean isAssignment() {
		return type == RelativeType.GOI;
	}

	/**
	 * Append restrictive facts to supplied vector with query objects
	 * @param resolver
	 * @param restrictives receives restrictive fact
	 * @param incidentals receives any incidental facts
	 * @param querySubject serves as placeholder for the subject
	 */
	public void resolveRestrictive(IResolver resolver, Vector<IFact> restrictives, Vector<IFact> incidentals, ISubject querySubject) {
		
		if(!isRestrictive())
			return;
		
		IRelation r = null;
		
		switch(type) {
		case GOI:			// performed assignment on process--no longer important
			return;
		case PE:
			r = resolver.getRelationByName("pe");							// use dummy relation
			restrictives.add(resolver.createFact(
					true, r,
					querySubject,
					subject.resolve(resolver, incidentals),
					null, null, null));
			break;
		case PO:
			r = resolver.getRelationByName("po");							// use dummy relation
			restrictives.add(resolver.createFact(
					true, r,
					querySubject,
					subject.resolve(resolver, incidentals),
					null, null, null));
			break;
		case POI:
			resolver.pushAssignment("ke'a", querySubject);							// ke'a resolves to a query
			restrictives.add(sentence.resolve(resolver, incidentals, true));		// resolve subsentence
			resolver.popAssignment("ke'a");											// clear ke'a
			break;
		case NOI:
			throw new AnalysisException("noi does not restrict--cannot resolve");
		default:
			throw new AnalysisException("unhandled relative type");
		}
	}

	/**
	 * Append incidental facts to supplied vector with query objects replaced by subject (for incidentals).
	 * @param resolver
	 * @param subject
	 * @param incidentalFacts
	 */
	public void resolveIncidental(IResolver resolver, ISubject subject, Vector<IFact> incidentalFacts) {
		if(type != RelativeType.NOI)
			throw new AnalysisException("only noi restricts -- cannot resolve");
		
		resolver.pushAssignment("ke'a", subject);						// ke'a resolves to parent subject
		Vector<IFact> v = new Vector<IFact>();							// collect incidentals
		incidentalFacts.add(sentence.resolve(resolver, v, true));		// resolve and add sentence fact
		incidentalFacts.addAll(v);										// also add any incidentals
		resolver.popAssignment("ke'a");									// clear ke'a
	}
	
	public String toString() {
		StringBuilder sb = new StringBuilder();
		
		switch(type) {
		case PE:
		case PO:
			sb.append(type + ":" + subject);
			break;
		case NOI:
		case POI:
			sb.append(type + ":" + sentence);
			break;
		case GOI:
			sb.append("{it-" + ReferenceAnalysis.getProsumtiNumber(prosumti) + "}");
			break;
		default:
			sb.append("#ERROR:RCA#");
		}
		
		return sb.toString();
	}

	/**
	 * @return render approximately into English
	 */
	public String toTranslation(AbstractDictionary ad) {
		StringBuilder sb = new StringBuilder();
		
		switch(type) {
		case PE:
			sb.append(ad.getWord("pe") + " " + subject.toTranslation(ad));
			break;
		case PO:
			sb.append(ad.getWord("po") + " " + subject.toTranslation(ad));
			break;
		case NOI:
			sb.append(ad.getWord("noi") + " " + sentence.toTranslation(ad));
			break;
		case POI:
			sb.append(ad.getWord("poi") + " " + sentence.toTranslation(ad));
			break;
		case GOI:
			sb.append("{" + ad.getWord("ko'a") + "-" + ReferenceAnalysis.getProsumtiNumber(prosumti) + "}");
			break;
		default:
			sb.append("#ERROR:RCA#");
		}
		
		return sb.toString().trim();
	}
	
	public String getProsumti() {
		return prosumti;
	}
}
