/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.analyzer;

import java.util.Vector;

import lojban.parser.GenericParseNode;
import lojban.parser.AbstractDictionary;

/**
 * @author mike
 * Analyzes the nodes which indicate a sentence (Bridi).
 */
class SentenceAnalysis extends AbstractRecursiveAnalysis {
	
	private TextAnalysis ta;
	
	/** if this is a top level sentence */
	private boolean topLevel = false;
	
	/** true if this sentence is being asked for confirmation */
	private boolean confirmQuery = false;
	
	/**
	 * Standard analyzer, ready for processing.
	 * @param ta analyzer
	 * @param topLevel true if this is a top level sentence
	 * @param queryNext true if this sentence is being asked with a xu phrase
	 */
	public SentenceAnalysis(TextAnalysis ta, boolean topLevel, boolean confirmQuery) {
		this.ta = ta;
		this.topLevel = topLevel;
		this.confirmQuery = confirmQuery;
	}
	
	/**
	 * Compound analyzer, for aggregating two already analyzed sentences, NOT FOR PARSING
	 */
	public SentenceAnalysis(SentenceAnalysis s1, SentenceAnalysis s2, IFact.Connective c, boolean n1, boolean n2, boolean rev, String externalText) {
		sen1 = s1;
		sen2 = s2;
		connective = c;
		neg1 = n1;
		neg2 = n2;
		external = true;
		extText = externalText;
		reverse = rev;
	}
	
	////// IN CASE OF PROBRIDI INVOCATION //////
	private SentenceAnalysis antecedent = null;

	////// FOR COMPLEX SENTENCES //////
	
	/** if this is an externally supplied analysis */
	private boolean external = true;
	
	/** the text corresponding to a connective, if external */
	private String extText = null;
	
	/** a connective term */
	private IFact.Connective connective = null;
	
	/** negations of connected subsentences and whether se reverse them */
	private boolean neg1 = false, neg2 = false, reverse = false;
	
	/** subsentences */
	private SentenceAnalysis sen1 = null, sen2 = null;
	
	/** resolved subjects */
	private ISubject[] resolved = new ISubject[5];
	
	/** true if next compound is a xu query */
	private boolean queryNext = false;
	
	////// FOR SIMPLE SENTENCES //////
	
	/** the sumti */
	private ReferenceAnalysis references[] = new ReferenceAnalysis[5];
	
	/** the selbri */
	private RelationAnalysis relation = null;
	
	/** override of next place by a fa clause */
	private int faLocation = 0;
	
	/** next place a reference goes in the bridi by natural order */
	private int nextLocation = 0;
	
	/** true if this bridi is a query on the selbri */ 
	private boolean querySelbri = false;
	
	/** if true, prosumti inverts meaning of antecedent */
	private boolean prosumtiNegate = false;
	
	/**
	 * Resolve this sentence to a fact.
	 * @param resolver resolver object
	 * @param incidentals receives incidental facts from this sentence
	 * @param if this is a relative clause--implies there must be a ke'a somewhere (x1 by default)
	 * @return the immediate fact of this sentence
	 */
	public IFact resolve(IResolver resolver, Vector<IFact> incidentals, boolean requireKeha) {
		IFact f = null;
		if(connective == null)		// delegate to simple or complex resolution
			f = simpleResolve(resolver, incidentals, requireKeha);
		else
			f = complexResolve(resolver, incidentals, requireKeha);
		return f;
	}
	
	public IFact resolve(IResolver resolver, Vector<IFact> incidentals) {
		return resolve(resolver, incidentals, false);
	}
		
	/**
	 * Resolve a simple sentence (just a bridi) to a fact
	 * @param resolver resolver object
	 * @param incidentals receives incidental facts from this sentence
	 * @return the immediate fact of this sentence
	 */
	private IFact simpleResolve(IResolver resolver, Vector<IFact> incidentals, boolean requireKeha) {
		
		// resolve relation
		IRelation r;
		if(querySelbri)
			r = resolver.getQueryRelation();
		else
			r = relation.resolve(resolver);
		
		// required ke' has been seen
		boolean kehaSeen = !requireKeha;
		
		// resolve all subjects and collect incidentals
		resolved = new ISubject[5];
		for(int i = 0; i < 5; i++)
			if(references[i] != null) {
				resolved[i] = references[i].resolve(resolver, incidentals);
				if(references[i].isKeha())
					kehaSeen = true;
			}
		
		// add ke'a in first empty spot if not seen
		if(!kehaSeen)
			for(int i = 0; i < 5; i++)
				if(resolved[i] == null) {
					resolved[i] = resolver.getAssignment("ke'a");
					break;
				}
		
		// combine resulting facts and return
		if(!confirmQuery) {
			return resolver.createFact(!relation.isNegated(), r,
					resolved[0], resolved[1], resolved[2], resolved[3], resolved[4]);
		} else {
			return resolver.createConfirmFact(!relation.isNegated(), r,
					resolved[0], resolved[1], resolved[2], resolved[3], resolved[4]);
		}
	}
	
	/**
	 * @return relation analysis
	 */
	private RelationAnalysis getRelation() {
		return relation;
	}

	/**
	 * Resolve a complex sentence (with ga...gi...) to a fact
	 * @param resolver resolver object
	 * @param incidentals receives incidental facts from this sentence
	 * @return the immediate fact of this sentence
	 */
	private IFact complexResolve(IResolver resolver, Vector<IFact> incidentals, boolean requireKeha) {
	
		// resolve the two sentences
		IFact f1 = sen1.resolve(resolver, incidentals, requireKeha);
		IFact f2 = sen2.resolve(resolver, incidentals, requireKeha);
		
		// consider that they may need to be reversed by a se clause
		if(!reverse) {
			return resolver.createConnectiveFact(connective, neg1, neg2, f1, f2);
		} else
			return resolver.createConnectiveFact(connective, neg2, neg1, f2, f1);
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.AbstractRecursiveAnalysis#getDebugName()
	 */
	public String getDebugName() {
		return "Sentence";
	}
	
	public void processSentence(GenericParseNode n) {
		processChildren(n);
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.AbstractRecursiveAnalysis#end()
	 */
	public void end() {
		
		if(connective == null) {
			
			// handle prosumti invocation
			if(relation.getInvocation() != null) {
				String pro = relation.getInvocation();
				prosumtiNegate = relation.isNegated();
				
				if(pro.equals("mo"))
					querySelbri = true;
				else {
					// get antecedent and relation
					antecedent = ta.getAssignedRel(pro);
					relation = new RelationAnalysis(antecedent.getRelation(), prosumtiNegate);
						
					// look for any extra subjects
					for(int i = 0; i < 5; i++)
						if(references[i] == null)
							references[i] = antecedent.getReference(i);
				}
			}
			
			// handle prosumti assignment
			if(relation.getAssignment() != null) {
				ta.makeAssignment(relation.getAssignment(), this);
			}
			
			// look for/resolve reflexive prosumti
			for(int i = 0; i < 5; i++) {
				if(references[i] != null && references[i].getReflexive() >= 0)	// is reflexive
					references[i].setAntecedent(references[references[i].getReflexive()]);
			}
			
			if(topLevel) {
				ta.makeAssignment("go'i", this);
			}
		}
	}
	
	public void processTerms(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processTerms1(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processTerms2(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processTerm(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processTerm1(GenericParseNode n) {		
		processChildren(n);
	}
	
	/**
	 * Delegate sumti processing to ReferenceAnalysis
	 * @param n
	 */
	public void processSumti(GenericParseNode n) {
		
		// get the reference
		ReferenceAnalysis refa = new ReferenceAnalysis(ta);
		refa.processSumti(n);
		
		// if a fa clause was seen, put it in a special place
		// ensure by assert we don't overwrite
		if(faLocation != 0) {
			assert references[faLocation-1] == null;
			references[faLocation-1] = refa;
			nextLocation = faLocation;		// fill after this place
			faLocation = 0;					// forget fa clause
			return;
		}
		
		// otherwise find first empty space left
		// ensure by assert we don't overwrite
		while(references[nextLocation] != null)
			nextLocation++;
		assert references[nextLocation] == null;
		references[nextLocation++] = refa;
		
	}
	
	/**
	 * Record the location signified by a fa clause.
	 * @param n
	 */
	public void processFaclause(GenericParseNode n) {
		
		String s = n.getText();
		String FA[] = new String[]{"fa","fe","fi","fo","fu"};
		for(int i = 0; i < 5; i++)
			if(s.equals(FA[i])) {
				faLocation = i+1;
				return;
			}
	}
	
	/**
	 * Safe to ignore -- only needed for parser.
	 * @param n
	 */
	public void processCuclause(GenericParseNode n) {
		
	}
	
	public void processBriditail(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processBriditail1(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processBriditail2(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processBriditail3(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Delegate selbri analysis to RelationAnalysis
	 * @param n
	 */
	public void processSelbri(GenericParseNode n) {
		
		// get relation (ensure not overwriting)
		assert relation == null;
		relation = new RelationAnalysis(ta);
		relation.process(n);
		
		// after seeing the relation, default place is x2, not x1
		if(nextLocation == 0)
			nextLocation++;
	}
	
	public void processTailterms(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processGeksentence(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Handle the Gek clause and all of its subclauses, signifying subsentences.
	 * @param n
	 */
	public void processGek(GenericParseNode n) {		
		for(GenericParseNode n1 : n.getChildren()) {		
			if(n1.getType().equals("GAClause")) {					// GAClause
				processChildren(n1);
			} else if(n1.getType().equals("SEClause")) {			// SEClause
				reverse = true;
			} else if(n1.getType().equals("NAIClause")) {			// NAIClause
				neg1 = true;
			} else {
				throw new AnalysisException("gek subclause unknown: " + n1.getType());
			}
		}
	}
	
	/**
	 * this may contain a xu
	 */
	public void processGapost(GenericParseNode n) {
		if(n.getText().equals("xu"))
			queryNext = true;
		else
			throw new AnalysisException("unknown gapost: " + n.getText());
	}
	
	/**
	 * this is where the logical connective part of the ga phrase will occur
	 */
	public void processGapre(GenericParseNode n) {
		switch(n.getText().charAt(1)) {
		case 'a':
			connective = IFact.Connective.OR;
			break;
		case 'e':
			connective = IFact.Connective.AND;
			break;
		case 'o':
			connective = IFact.Connective.IFF;
			break;
		case 'u':
			connective = IFact.Connective.WON;
			break;
		default:
			throw new AnalysisException("unknown GA: " + n.getText());
		}
	}
	
	/**
	 * Handle Gik (which just determines whether second half is negated)
	 * @param n
	 */
	public void processGik(GenericParseNode n) {
		neg2 = n.getText().indexOf("nai") > 0;
		queryNext = n.getText().indexOf("xu") > 0;
	}
	
	/**
	 * Delegate subsentence, as this is a compound sentence
	 * @param n
	 */
	public void processSubsentence(GenericParseNode n) {
		SentenceAnalysis s = new SentenceAnalysis(ta, topLevel, queryNext);	// delegate analysis
		s.process(n.getChildren()[0]);
		
		queryNext = false;
		
		if(sen1 == null)		// this is the first part
			sen1 = s;
		else					// this is the second part
			sen2 = s;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		
		if(connective == null) {
		
			// SIMPLE SENTENCE
			
			StringBuilder sb = new StringBuilder();
			if(relation.isNegated())
				sb.append("~");
			
			// relation
			if(querySelbri)
				sb.append("?(");
			else {
				sb.append(relation);
				if(confirmQuery)
					sb.append("?");
				sb.append("(");
			}
			
			// subjects
			for(int i = 0; i < 5; i++)
				if(references[i] == null) {
					sb.append("-, ");
				} else
					sb.append(references[i] + ", ");
			sb.delete(sb.length()-2, sb.length());	// remove extra commas
			sb.append(")");
			return sb.toString();
			
		} else {
			
			// COMPOUND SENTENCE
			StringBuilder sb = new StringBuilder();
			
			// connective
			switch(connective) {
				case OR: sb.append("or("); break;
				case IFF: sb.append("iff("); break;
				case AND: sb.append("and("); break;
				case WON: sb.append("won("); break;
			}
			
			// first sentence
			if(!reverse && neg1 || reverse && neg2)
				sb.append("~");
			if(!reverse)
				sb.append(sen1);
			else
				sb.append(sen2);
			sb.append(", ");
		
			// second sentence
			if(!reverse && neg2 || reverse && neg1)
				sb.append("~");
			if(!reverse)
				sb.append(sen2);
			else
				sb.append(sen1);
			sb.append(")");
			
			return sb.toString();
			
		}
	}
	
	/**
	 * Get a reference analysis.
	 * @param i 0-4
	 * @return the reference analysis
	 */
	private ReferenceAnalysis getReference(int i) {
		return references[i];
	}

	/**
	 * @return approximation in psuedo-English
	 */
	public String toTranslation(AbstractDictionary ad) {
		
		if(connective == null) {
		
			// SIMPLE SENTENCE
			StringBuilder sb = new StringBuilder();
			
			// first subject
			if(references[0] == null) {
				sb.append(ad.getWord("?") + " ");
			} 						// x1 place
			else {
				sb.append(references[0].toTranslation(ad).trim() + " ");
			}
			
			if(confirmQuery)
				sb.append(ad.getWord("xu") + " ");
			
			// relation itself
			if(querySelbri)
				sb.append(ad.getWord("mo") + " ");
			else {
				if(relation.isNegated())						// relation itself
					sb.append(ad.getWord("na") + " ");
				sb.append(ad.getVerb(relation.toString()).trim() + " ");
			}
			
			// second through fifth subjects
			for(int i = 1; i < 5; i++)						// x2-x5 places
				if(references[i] != null) {
					sb.append(ad.getSubjectTag(relation.toString(), i+1).trim() + " ");
					sb.append(references[i].toTranslation(ad).trim() + " ");
				}
			return sb.toString().trim();
		
		} else {
			
			// COMPOUND SENTENCE
			StringBuilder sb = new StringBuilder();
			
			// first sentence
			sb.append("(");
			if(!reverse)
				sb.append(sen1.toTranslation(ad));
			else
				sb.append(sen2.toTranslation(ad));
			sb.append(") ");

			// connective
			if(!reverse && neg1 || reverse && neg2)
				sb.append(ad.getWord("na") + "-");
			switch(connective) {
				case OR: sb.append(ad.getWord("a")); break;
				case IFF: sb.append(ad.getWord("o")); break;
				case AND: sb.append(ad.getWord("e")); break;
				case WON: sb.append(ad.getWord("u")); break;
			}
			if(!reverse && neg2 || reverse && neg1)
				sb.append("-" + ad.getWord("na"));
			
			// second sentence
			sb.append(" (");
			if(!reverse)
				sb.append(sen2.toTranslation(ad));
			else
				sb.append(sen1.toTranslation(ad));
			sb.append(")");
			
			return sb.toString();
			
		}
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.AbstractRecursiveAnalysis#getText()
	 */
	public String getText() {
		if(!external)
			return super.getText();
		else {
			return extText;
		}
	}
}
