/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.analyzer;

import java.util.LinkedList;
import java.util.TreeMap;
import java.util.Vector;

import lojban.parser.AbstractDictionary;
import lojban.parser.GenericParseNode;

/**
 * @author mike
 * Analyzes the nodes which make up a full utterance.
 */
class TextAnalysis extends AbstractRecursiveAnalysis {
	
	private Vector<SentenceAnalysis> sentences = new Vector<SentenceAnalysis>();
	
	/** information for a connective phrase */
	private IFact.Connective connective = null;
	private boolean neg1 = false, neg2 = false;
	private boolean reverse = false;

	/** set true if the next sentence is of uncertain truth value */
	private boolean queryNext = false;
	
	private GenericParseNode connectNode = null;
	
	/** assignable prosumti/bridi maps */
	private TreeMap<String, ReferenceAnalysis> assignableProsumti = new TreeMap<String, ReferenceAnalysis>();
	private TreeMap<String, SentenceAnalysis> assignableProbridi = new TreeMap<String, SentenceAnalysis>();
	
	/** sumti list, for finding anaphoric references */
	private LinkedList<ReferenceAnalysis> sumti = new LinkedList<ReferenceAnalysis>();

	/**
	 * Copy over persistent data from old text analysis
	 * @param ta
	 */
	public TextAnalysis(TextAnalysis ta) {
		if(ta == null)
			return;
		else {
			assignableProsumti = ta.assignableProsumti;
			assignableProbridi = ta.assignableProbridi;
			sumti = ta.sumti;
		}
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.AbstractRecursiveAnalysis#getDebugName()
	 */
	public String getDebugName() {
		return "Text";
	}
	
	public void processText(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processText1(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processParagraphs(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processParagraph(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processStatement(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processStatement1(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processStatement2(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processStatement3(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processFragment(GenericParseNode n) {
		throw new AnalysisException("parse error/incomplete sentence (missing terminator?): " + n.getText());
	}
	
	public void processIclause(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Ignore, .i only needed for parser
	 */
	public void processIpre(GenericParseNode n) {
	}
	
	public void processIpost(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processPostclause(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processIndicators(GenericParseNode n) {
		queryNext = true;
	}
	
	public void processTextpart2(GenericParseNode n) {
		processChildren(n);
	}
	
	public void processJoikjek(GenericParseNode n) {
		processChildren(n);
	}
	
	/**
	 * Handle a connective between sentences.  Save it in
	 * private variables for use next time we finish a sentence.
	 * @param n
	 */
	public void processJek(GenericParseNode n) {
		
		connectNode = n;
		
		for(GenericParseNode n1 : n.getChildren()) {
			
			String type = n1.getType();
			if(type.equals("JAClause")) {
				switch(n1.getText().charAt(1)) {
				case 'a':
					connective = IFact.Connective.OR;
					break;
				case 'e':
					connective = IFact.Connective.AND;
					break;
				case 'o':
					connective = IFact.Connective.IFF;
					break;
				case 'u':
					connective = IFact.Connective.WON;
					break;
				default:
					throw new AnalysisException("unknown JA: " + n1.getText());
				}
			} else if(type.equals("NAIClause")) {
				neg2 = true;
			} else if(type.equals("NAClause")) {
				neg1 = true;
			} else if(type.equals("SEClause")) {
				reverse = true;
			} else {
				throw new AnalysisException("unknown jek subclause: " + type);
			}
			
		}
	}
	
	
	/**
	 * Delegate sentence parsing to SentenceAnalysis
	 * @param n
	 */
	public void processSentence(GenericParseNode n) {
		
		// hand over to sentence analysis
		SentenceAnalysis sa = new SentenceAnalysis(this, true, queryNext);
		sa.process(n);
		
		if(connective != null) {		// COMPOUND SENTENCE
			
			// this is better handled at the sentence analysis level,
			// therefore we delegate to a special sentence analysis class which
			// is generated in place and without any nodes to process
			
			// get back last sentence
			SentenceAnalysis left = sentences.get(sentences.size()-1);
			
			// make new sentence that aggregates previous two
			SentenceAnalysis compound = new SentenceAnalysis(
					left, sa, connective, neg1, neg2, reverse,
					left.getText() + " " + connectNode.getText() + " " + sa.getText());
			
			// replace last sentence with compound sentence
			sentences.set(sentences.size()-1, compound);
			
			// reset connective variables
			connectNode = null;
			connective = null;
			neg1 = neg2 = reverse = false;
		} else {						// SIMPLE SENTENCE
			sentences.add(sa);
		}
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for(SentenceAnalysis sa : sentences)
			sb.append(sa + "; ");
		if(sentences.size() > 0)
			sb.delete(sb.length()-2, sb.length());	// remove extra commas
		return sb.toString();
	}
	
	/**
	 * Generate an approximate translation.
	 * @param ad dictionary for target language
	 * @return approximate translation
	 */
	public String toTranslation(AbstractDictionary ad) {
		StringBuilder sb = new StringBuilder();
		for(SentenceAnalysis sa : sentences) {
			String s = sa.toTranslation(ad);
			sb.append(Character.toUpperCase(s.charAt(0)));
			sb.append(s.substring(1) + ".  ");
		}
		return sb.toString();
	}
	
	/**
	 * Resolve text to a list of facts
	 * @param resolver
	 * @return facts extracted from text
	 */
	public IFact[] resolve(IResolver resolver) {
		Vector<IFact> facts = new Vector<IFact>();
		
		// collect facts and their incidentals
		for(SentenceAnalysis sa : sentences) {
			Vector<IFact> incidentals = new Vector<IFact>();
			facts.add(sa.resolve(resolver, incidentals));
			facts.addAll(incidentals);
		}
		
		return facts.toArray(new IFact[]{});
	}

	/**
	 * Make an assignment of a prosumti to a reference analysis
	 * @param prosumti
	 * @param analysis
	 */
	public void makeAssignment(String prosumti, ReferenceAnalysis analysis) {
		assignableProsumti.put(prosumti, analysis);
	}
	
	/**
	 * Make an assignment of a probridi to a sentence analysis
	 * @param prosumti
	 * @param analysis
	 */
	public void makeAssignment(String probridi, SentenceAnalysis analysis) {
		assignableProbridi.put(probridi, analysis);
	}

	/**
	 * Get the reference assigned to a prosumti.
	 * @param prosumti
	 * @return the reference analysis
	 */
	public ReferenceAnalysis getAssignedRef(String prosumti) {
		if(assignableProsumti.containsKey(prosumti))
			return assignableProsumti.get(prosumti);
		else
			throw new AnalysisException("prosumti without antecedent: " + prosumti);
	}
	
	/**
	 * Get the reference assigned to a probridi.
	 * @param prosumti
	 * @return the sentence analysis
	 */
	public SentenceAnalysis getAssignedRel(String probridi) {
		if(assignableProbridi.containsKey(probridi))
			return assignableProbridi.get(probridi);
		else
			throw new AnalysisException("probridi without antecedent: " + probridi);
	}

	/**
	 * @param analysis
	 */
	public void recordSumti(ReferenceAnalysis ra) {
		sumti.addFirst(ra);
	}

	/**
	 * Get the last complete sumti which is legal for anaphoric reference.
	 * @return the antecedent of an anaphoric
	 */
	public ReferenceAnalysis getAnaphoricRef() {
		for(ReferenceAnalysis ra : sumti)
			if(ra.isComplete() && ra.isAnaphoricAntecedent())
				return ra;
		throw new AnalysisException("no legal anaphoric antecedent");
	}
}
