/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.gui;

import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JEditorPane;
import javax.swing.JSeparator;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.FileFilter;

import lojban.agent.IConversationListener;
import lojban.agent.LojbanContext;
import lojban.agent.LojbanKnowledgeBase;
import lojban.analyzer.Analysis;
import lojban.analyzer.AnalysisException;
import lojban.parser.EnglishDictionary;
import lojban.parser.ParseException;
import lojban.parser.SpanishDictionary;

public class LojbanGui extends JFrame implements IConversationListener, ActionListener {
	private static final long serialVersionUID = 1L;

	private LojbanContext lc = null;
	
	private static final String header = "<HTML><TABLE>";
	private static final String footer = "</TABLE></HTML>";
	private static final String error = "<TR><TD VALIGN=TOP ALIGN=CENTER><FONT COLOR=RED SIZE=+2>!</FONT></TD><TD>%s</TD></TR>";
	private static final String input = "<TR><TD VALIGN=TOP ALIGN=CENTER><FONT COLOR=GREEN SIZE=+2>&gt;</FONT></TD><TD>%s</TD></TR>";
	private static final String output = "<TR><TD VALIGN=TOP ALIGN=CENTER><FONT COLOR=BLUE SIZE=+2>*</FONT></TD><TD><B>%s</B></TD></TR>";
	private static final String translation = "<TR><TD></TD><TABLE><TR><TD VALIGN=TOP ALIGN=CENTER><FONT COLOR=GRAY><B>%l</B></FONT></TD><TD><I><FONT COLOR=GRAY SIZE=-1>%s</FONT></I></TD></TR></TABLE></TD></TR>";
	
	private StringBuilder content = new StringBuilder();
	
	private JEditorPane text = new JEditorPane("text/html", header + footer);
	private JScrollPane main = new JScrollPane(text);
	private JTextField entry = new JTextField();
	
	private JButton abort = new JButton("    x    ");
	private JLabel statusText = new JLabel("Ready.");

	private JMenuBar bar = null;
	private JMenuItem load = new JMenuItem("Load");
	private JMenuItem clear = new JMenuItem("Clear");
	private JMenuItem quit = new JMenuItem("Quit");
	private JMenuItem english = new JCheckBoxMenuItem("English");
	private JMenuItem spanish = new JCheckBoxMenuItem("Spanish");
	private JMenuItem multiple = new JCheckBoxMenuItem("Multiple Answers");
	private JMenuItem about = new JMenuItem("About");
	private JMenu pluginMenu = new JMenu("Plugins");
	
	private Thread processThread = null;
	
	/** for translations */
	private Analysis responseAnalysis = new Analysis();
	private Analysis inputAnalysis = new Analysis();
		
	/**
	 * Build GUI around a context.
	 * @param lc
	 */
	public LojbanGui(LojbanContext lc) {
		super("Lojban Reasoning Engine Context");
		setSize(640, 480);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		// save context
		this.lc = lc;
		lc.addConversationListener(this);
		
		// configure statusbar
		JPanel statusbar = new JPanel();
		abort.setMargin(new Insets(0, 0, 0, 0));
		abort.setEnabled(false);
		abort.addActionListener(this);
		statusbar.setLayout(new BorderLayout());
		statusbar.add(statusText, BorderLayout.CENTER);
		statusbar.add(abort, BorderLayout.EAST);
		
		// configure lower panel
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.add(entry, BorderLayout.CENTER);
		panel.add(statusbar, BorderLayout.SOUTH);
		
		// configure main window
		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(main, BorderLayout.CENTER);
		getContentPane().add(panel, BorderLayout.SOUTH);
		text.setEditable(false);
		entry.setFont(new Font("Times New Roman", Font.BOLD, 16));
		entry.addActionListener(this);
		
		// configure menu
		bar = new JMenuBar();
		this.setJMenuBar(bar);
		JMenu menu = new JMenu("Conversation");
		bar.add(menu);
		load.addActionListener(this);
		about.addActionListener(this);
		clear.addActionListener(this);
		quit.addActionListener(this);
		multiple.addActionListener(this);
		menu.add(load);
		menu.add(clear);
		menu.add(new JSeparator());
		menu.add(multiple);
		menu.add(new JSeparator());
		menu.add(quit);
		menu = new JMenu("Translation");
		bar.add(menu);
		menu.add(english);
		menu.add(spanish);
		bar.add(pluginMenu);
		menu = new JMenu("Help");
		menu.add(about);
		bar.add(menu);
		refreshPlugins();
		
		// run
		setVisible(true);
	}
	
	/**
	 * Update the plugin list in the menu.
	 */
	private void refreshPlugins() {
		String[] s = lc.listPlugins();
		pluginMenu.removeAll();
		for(String s1 : s) {
			JCheckBoxMenuItem m = new JCheckBoxMenuItem(s1);
			m.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					JCheckBoxMenuItem jcbmi = (JCheckBoxMenuItem) ae.getSource();
					lc.setPluginStatus(jcbmi.getText(), jcbmi.getState());
				}		
			});
			m.setState(lc.getPluginStatus(s1));
			pluginMenu.add(m);
		}
	}
	
	/* (non-Javadoc)
	 * @see lojban.agent.IConversationListener#input(java.lang.String)
	 */
	public void input(String s) {
		content.append(input.replace("%s", s));
				
		doTranslation(s, false);
		try {
			responseAnalysis.analyze(s);	// also track responses for probridi assignment
		}
		catch(AnalysisException e) { }	// ignore failures -- from parse errors anyways, so it didn't make the kb
		catch(ParseException e) { }
		
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				text.setText(content.toString());
				main.getVerticalScrollBar().setValue(main.getVerticalScrollBar().getMaximum());
			}
		});
	}

	/* (non-Javadoc)
	 * @see lojban.agent.IConversationListener#output(java.lang.String)
	 */
	public void output(String s) {
		content.append(output.replace("%s", s));
		
		doTranslation(s, true);
			
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				text.setText(content.toString());
				main.getVerticalScrollBar().setValue(main.getVerticalScrollBar().getMaximum());
			}
		});
	}
	
	/**
	 * Add a translation line.
	 */
	protected void doTranslation(String s, boolean response) {
		Analysis translationAnalysis = null;
		if(response)
			translationAnalysis = responseAnalysis;
		else
			translationAnalysis = inputAnalysis;
		if(english.isSelected() || spanish.isSelected()) {
			
			try {
				
				translationAnalysis.analyze(s);
				
				if(english.isSelected())
					content.append(translation.replace("%l", "en").replace("%s",
						translationAnalysis.toTranslation(new EnglishDictionary())));
				if(spanish.isSelected())
					content.append(translation.replace("%l", "es").replace("%s",
						translationAnalysis.toTranslation(new SpanishDictionary()).replace("^", "")));
			}
			catch(AnalysisException e) { }	// ignore failures -- from parse errors anyways, so it didn't make the kb
			catch(ParseException e) { }
		}
	}

	/* (non-Javadoc)
	 * @see lojban.agent.IConversationListener#error(java.lang.String)
	 */
	public void error(String s) {
		if(s != null)
			content.append(error.replace("%s", s));
		else
			content.append(error.replace("%s", "unknown error"));
		text.setText(content.toString());
	}

	/* (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent ae) {
		if(ae.getSource() == entry) {
			entry.setEnabled(false);
			bar.setEnabled(false);
			abort.setEnabled(true);
			processThread = new Thread() {
				public void run() {
					lc.speak(entry.getText());
					SwingUtilities.invokeLater( new Runnable() {
						public void run() {
							entry.setText("");
							entry.setEnabled(true);
							entry.requestFocus();
							bar.setEnabled(true);
							abort.setEnabled(false);
							processThread = null;
						}
					});
				}
			};
			processThread.start();
		}
		if(ae.getSource() == load) {
			//open a file-open dialog
			//read the file in line by line
			final JFileChooser chooser = new JFileChooser();
			FileFilter myFilter = new FileFilter(){
				
				String getExtension(File f) {
					if(f != null) {
					    String filename = f.getName();
					    int i = filename.lastIndexOf('.');
					    if(i>0 && i<filename.length()-1) {
						return filename.substring(i+1).toLowerCase();
					    };
					}
					return null;
				}
					
				public boolean accept(File arg0) {
					if(arg0.isDirectory())
						return true;
					if("lojban".equals(getExtension(arg0)))
						return true;
					else
						return false;
				}

				@Override
				public String getDescription() {
					return "Lojban files";
				}
				
			};
			chooser.setFileFilter(myFilter);
			int returnVal = chooser.showOpenDialog(this);
		    if(returnVal == JFileChooser.APPROVE_OPTION) {
		    		
	    		entry.setEnabled(false);
				bar.setEnabled(false);
				abort.setEnabled(true);
				processThread = new Thread() {
					public void run() {
						
						try {
							BufferedReader bufferedReader = new BufferedReader(new FileReader(chooser.getSelectedFile()));
							String line=null;
							while((line = bufferedReader.readLine()) !=null){
								if(line.indexOf("#") >= 0)
									line = line.substring(0, line.indexOf("#")).trim();
								if(line != null && line.length() > 0)
									lc.speak(line);
							}
						} catch (FileNotFoundException e) {
							// This should never happen
							e.printStackTrace();
						} catch (IOException e) {
							// This should never happen either!
							e.printStackTrace();
						}
						
						SwingUtilities.invokeLater( new Runnable() {
							public void run() {
								entry.setEnabled(true);
								entry.requestFocus();
								bar.setEnabled(true);
								abort.setEnabled(false);
								processThread = null;
							}
						});
					}
				};
				processThread.start();
		    	
		    }
		}
		if(ae.getSource() == quit) {
			System.exit(0);
		}
		if(ae.getSource() == clear) {
			content.delete(0, content.length());
			text.setText(content.toString());
		}
		if(ae.getSource() == abort) {
			processThread.interrupt();
			entry.setEnabled(true);
			entry.requestFocus();
			bar.setEnabled(true);
			abort.setEnabled(false);
			processThread = null;
			statusText.setText("Killed.  Warning: KB may not be stable.");
		}
		if(ae.getSource() == multiple) {
			lc.setMultiple(multiple.isSelected());
		}
		if(ae.getSource()==about){
			new About(this);
		}
	}
	
	public void status(final String s) {
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				statusText.setText(s);
			}
		});
	}
	
	/**
	 * Get a context and show GUI.
	 * @param args
	 */
	public static void main(String args[]) {
		LojbanContext lc = new LojbanKnowledgeBase().getContext();
		lc.addPlugin(new StdOutPlugin());
		lc.addPlugin(new TextToSpeechPlugin());
		new LojbanGui(lc);
	}
}
