/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.kb.jtp;

import java.util.LinkedList;

import lojban.agent.KnowledgeBaseException;
import lojban.analyzer.AnalysisException;
import lojban.analyzer.IFact;
import lojban.analyzer.IRelation;
import lojban.analyzer.ISubject;

public class JTPConnectiveFact extends JTPFact {

	private boolean neg1, neg2;
	private JTPFact f1, f2;
	private Connective c = null;
	
	public JTPConnectiveFact(Connective c, boolean neg1, boolean neg2, JTPFact fact1, JTPFact fact2) {
		this.c = c;
		this.neg1 = neg1;
		this.neg2 = neg2;
		f1 = fact1;
		f2 = fact2;
		
		if(f1 == null)
			throw new KnowledgeBaseException("connective with null left child");
		if(f2 == null)
			throw new KnowledgeBaseException("connective with null right child");
	}
	
	public boolean equals(Object o){
		if(o instanceof JTPConnectiveFact){
			return(((JTPConnectiveFact) o).c.equals(this.c) && ((JTPConnectiveFact) o).f1.equals(this.f1) && ((JTPConnectiveFact) o).f2.equals(this.f2) && this.neg1==((JTPConnectiveFact) o).neg1 && this.neg1==((JTPConnectiveFact) o).neg2);
		}else
			return false;
	}
	
	public int hashCode(){
		return (f1.hashCode() ^ f2.hashCode() & c.hashCode() + (neg1?1:0) + (neg2?1:0));
	}
	
	public String toInnerQuery(boolean soft) {
		StringBuilder sb = new StringBuilder();
		switch(c) {
		case OR: sb.append("(or "); break;
		case IFF: sb.append("(<=> "); break;
		case WON: sb.append("(won "); break;
		case AND: sb.append("(and "); break;
		}
		if(neg1)
			sb.append("(not ");
		sb.append(f1.toInnerQuery(soft));
		if(neg1)
			sb.append(")");
		sb.append(" ");
		if(neg2)
			sb.append("(not ");
		sb.append(f2.toInnerQuery(soft));
		if(neg2)
			sb.append(")");
		sb.append(")");
		return sb.toString();
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return toInnerQuery(false);
	}

	/* (non-Javadoc)
	 * @see lojban.kb.jtp2.JTPFact#getLateSubjects()
	 */
	public JTPLateSubject[] getLateSubjects() {
		LinkedList<JTPLateSubject> subjs = new LinkedList<JTPLateSubject>();
		for(JTPLateSubject s : f1.getLateSubjects())
			if(!subjs.contains(s))
				subjs.add(s);
		for(JTPLateSubject s : f2.getLateSubjects())
			if(!subjs.contains(s))
				subjs.add(s);
		return subjs.toArray(new JTPLateSubject[]{});
	}

	/* (non-Javadoc)
	 * @see lojban.kb.jtp2.JTPFact#getConcreteSubjects()
	 */
	public JTPConcreteSubject[] getConcreteSubjects() {
		LinkedList<JTPConcreteSubject> subjs = new LinkedList<JTPConcreteSubject>();
		for(JTPConcreteSubject s : f1.getConcreteSubjects())
			if(!subjs.contains(s))
				subjs.add(s);
		for(JTPConcreteSubject s : f2.getConcreteSubjects())
			if(!subjs.contains(s))
				subjs.add(s);
		return subjs.toArray(new JTPConcreteSubject[]{});
	}

	/* (non-Javadoc)
	 * @see lojban.kb.jtp2.JTPFact#substitute(java.lang.String, lojban.kb.jtp2.JTPSubject)
	 */
	public JTPFact substitute(JTPSubject id, JTPSubject s) {
		return new JTPConnectiveFact(c, neg1, neg2, f1.substitute(id, s), f2.substitute(id, s));		
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#getConnective()
	 */
	public Connective getConnective() throws AnalysisException {
		return c;
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#getSubject(int)
	 */
	public ISubject getSubject(int n) throws AnalysisException {
		throw new AnalysisException("Cannot get the subjects of a connective fact");
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#getRelation()
	 */
	public IRelation getRelation() throws AnalysisException {
		throw new AnalysisException("Cannot get the relations of a connective fact");
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#isComposite()
	 */
	public boolean isComposite() {
		return true;
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#getNegLeft()
	 */
	public boolean getNegLeft() {
		return neg1;
	}
	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#getNegRight()
	 */
	public boolean getNegRight() {
		return neg2;
	}
	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#getLeft()
	 */
	public IFact getLeft() {
		return f1;
	}
	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#getRight()
	 */
	public IFact getRight() {
		return f2;
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#isQuery()
	 */
	public boolean isQuery() {
		return f1.isQuery() || f2.isQuery();
	}

	public boolean isConfirmQuery() {
		return f1.isConfirmQuery() || f2.isConfirmQuery();
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IFact#isTrue()
	 */
	public boolean isTrue() throws AnalysisException {
		throw new AnalysisException("Cannot get the truth value of a connective fact");
	}
}
