/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.kb.jtp;

import lojban.analyzer.AnalysisException;
import lojban.analyzer.IFact;
import lojban.analyzer.ISubject;

public abstract class JTPFact implements IFact {

	public FactType getFactType() {
		if(toQuery().contains("?ask"))
			return IFact.FactType.QUERY;
		return IFact.FactType.STATEMENT;
	}
	
	public IFact substitute(ISubject id, ISubject s) {
		return substitute((JTPSubject)id, (JTPSubject)s);
	}
		
	/**
	 * Substitute any identifier subject with a replacement
	 * @param id original id
	 * @param s replacement subject
	 */
	public abstract JTPFact substitute(JTPSubject id, JTPSubject s);
	
	/**
	 * @return subjects in this fact that need clarification/resolution
	 */
	public abstract JTPLateSubject[] getLateSubjects();
	
	/**
	 * @return subjects in this fact that need concrete definitions
	 */
	public abstract JTPConcreteSubject[] getConcreteSubjects();
	
	public abstract String toInnerQuery(boolean soft);
	
	/**
	 * @return query string for this fact
	 */
	public String toQuery() {
		StringBuilder sb = new StringBuilder();
		JTPLateSubject lates[] = getLateSubjects();
		JTPConcreteSubject concretes[] = getConcreteSubjects();
		
		// get the query string
		// this is the result of this JTP query
		String queryString = toInnerQuery(false);
		
		// open the concrete enclosure
		// this will include objects that are instantiated here
		if(concretes.length > 0)
			sb.append("(and ");
		
		// open lates enclosure/clarify all late subjects
		// this will include things which exist already
		if(lates.length > 0) {
			if(!isQuery() && !isConfirmQuery())
				sb.append("(=> ");				// if it's a statement, it's an implication
			else
				sb.append("(and ");				// if it's a question, use and instead (don't want a false hypothesis!)
			
			sb.append("(and");
			for(JTPLateSubject late : lates)
				sb.append(" " + late.toQuery());
			sb.append(") ");
		}
		
		// add the query in
		sb.append(queryString);
						
		// close the lates enclosure
		if(lates.length > 0)
			sb.append(")");
		
		// add concrete definitions
		for(JTPConcreteSubject concrete : concretes)
			sb.append(" " + concrete.toQuery());
		
		// close the concrete enclosure
		if(concretes.length > 0)
			sb.append(")");
		
		return sb.toString();
	}

	public boolean getNegLeft() {
		throw new AnalysisException("called getNegLeft on simple fact");
	}
	public boolean getNegRight() {
		throw new AnalysisException("called getNegRight on simple fact");
	}
	public IFact getLeft() {
		throw new AnalysisException("called getLeft on simple fact");
	}
	public IFact getRight() {
		throw new AnalysisException("called getRight on simple fact");
	}
}
