/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.kb.jtp;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;

import jtp.ReasoningException;
import jtp.ReasoningStep;
import jtp.ReasoningStepIterator;
import jtp.context.BasicReasoningContext;
import jtp.disp.DispatcherUtils;
import jtp.fol.Literal;
import jtp.fol.Variable;
import jtp.func.Less;
import jtp.gmp.ClauseOrientationKB;
import jtp.modelim.AskingQueryProcessor;

import lojban.agent.KnowledgeBaseException;
import lojban.analyzer.IFact;
import lojban.analyzer.IKnowledgeBase;
import lojban.analyzer.IRelation;
import lojban.analyzer.IResolver;
import lojban.analyzer.ISubject;
import lojban.analyzer.IFact.Connective;

public class JTPResolver implements IResolver, IKnowledgeBase {

	/** the JTP Context */
	private BasicReasoningContext context;
	
	/** A KB for FOL clauses */
	private ClauseOrientationKB ckb;
	
	/** count for variable names */
	private int queryCount = 0;
	
	/** index names */
	private Set<String> names = new TreeSet<String>();
	
	/** index relations */
	private Map<String, JTPRelation> relations = new HashMap<String, JTPRelation>();
	
	private JTPLiteral MI = new JTPLiteral("mi", true);
	
	private JTPLiteral DO = new JTPLiteral("do", true);
	
	public TreeMap<String, Stack<JTPSubject>> assignments = new TreeMap<String, Stack<JTPSubject>>();
		
	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#createFact(boolean, lojban.analyzer.IRelation, lojban.analyzer.ISubject, lojban.analyzer.ISubject, lojban.analyzer.ISubject, lojban.analyzer.ISubject, lojban.analyzer.ISubject)
	 */
	public IFact createFact(boolean isTrue, IRelation r, ISubject s1, ISubject s2, ISubject s3, ISubject s4, ISubject s5) {
		return new JTPSimpleFact(isTrue, (JTPRelation)r, (JTPSubject)s1, (JTPSubject)s2, (JTPSubject)s3, (JTPSubject)s4, (JTPSubject)s5);
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#createConfirmFact(boolean, lojban.analyzer.IRelation, lojban.analyzer.ISubject, lojban.analyzer.ISubject, lojban.analyzer.ISubject, lojban.analyzer.ISubject, lojban.analyzer.ISubject)
	 */
	public IFact createConfirmFact(boolean isTrue, IRelation r, ISubject s1, ISubject s2, ISubject s3, ISubject s4, ISubject s5) {
		return new JTPSimpleFact(isTrue, (JTPRelation)r, (JTPSubject)s1, (JTPSubject)s2, (JTPSubject)s3, (JTPSubject)s4, (JTPSubject)s5, true);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#createConnectiveFact(lojban.analyzer.IFact.Connective, boolean, boolean, lojban.analyzer.IFact, lojban.analyzer.IFact)
	 */
	public IFact createConnectiveFact(Connective c, boolean neg1, boolean neg2, IFact fact1, IFact fact2) {
		return new JTPConnectiveFact(c, neg1, neg2, (JTPFact)fact1, (JTPFact)fact2);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#getRelationByName(java.lang.String)
	 */
	public IRelation getRelationByName(String name) {
		IRelation r = new JTPRelation(name);
		relations.put(name, (JTPRelation) r);
		return r;
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#getSubjectByName(java.lang.String)
	 */
	public ISubject getSubjectByName(String name, IFact[] f, ISubject q) {
				
		// append name fact
		JTPFact f1[] = new JTPFact[f.length+1];
		f1[0] = new JTPSimpleFact(true, (JTPRelation)getRelationByName("cmene"), (JTPSubject)getLiteralSubject(name), (JTPSubject)q, null, null, null);
		
		// the first use of the name is considered veridical, future uses are considered nonveridical
		// note that there may be a better way to handle this though this works for the preset
		boolean concrete;				// veridical
		if(names.contains(name)) {		// repeat usage, consider non-veridical
			concrete = false;
			
		} else {						// first usage, make it veridical
			concrete = true;
			names.add(name);
		}			
		
		// copy old facts
		for(int i = 0; i < f.length; i++)
			f1[i+1] = (JTPFact)f[i];
		
		// delegate
		return getSubjectByRestrictions(null, f1, q, concrete);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#getSubjectByRestrictions(lojban.analyzer.IRelation, lojban.analyzer.IFact[], lojban.analyzer.ISubject, boolean)
	 */
	public ISubject getSubjectByRestrictions(IRelation r, IFact[] f, ISubject q, boolean veridical) {
		
		// convert to a single query
		JTPFact query = null;
		if(r != null) {		
			query = (JTPFact)createFact(true, r, q, null, null, null, null);
		}
		for(IFact g1 : f) {
			if(query == null)
				query = (JTPFact) g1;
			else {
				query = (JTPFact) createConnectiveFact(Connective.AND, false, false, query, (JTPFact)g1);
			}
		}
				
		ISubject s;
		if(veridical)
			s = new JTPConcreteSubject(query, (JTPLiteral)q);
		else
			s = new JTPLateSubject(query, (JTPLiteral)q);
		
		//subjects.put(((JTPLiteral)q).toString(), (JTPSubject)s);
		
		return s;
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#getQuerySubject()
	 */
	public ISubject getRelativeSubject() {
		return new JTPLiteral("q" + queryCount++, false);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#getLiteralSubject(java.lang.String)
	 */
	public ISubject getLiteralSubject(String literal) {
		return new JTPLiteral("\"" + literal + "\"", true);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#pushAssignment(java.lang.String, lojban.analyzer.ISubject)
	 */
	public void pushAssignment(String prosumti, ISubject s) {
		Stack<JTPSubject> stack;
		if(assignments.containsKey(prosumti))
			 stack = assignments.get(prosumti);
		else
			assignments.put(prosumti, stack = new Stack<JTPSubject>());
		stack.push((JTPSubject)s);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#popAssignment(java.lang.String)
	 */
	public void popAssignment(String prosumti) {
		if(assignments.containsKey(prosumti)) {
			Stack<JTPSubject> stack = assignments.get(prosumti);
			if(stack.size() > 0)
				stack.pop();
			else
				throw new RuntimeException("empty assignment stack for " + prosumti);
		} else
			throw new RuntimeException("no assignment stack for " + prosumti);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#getAssignment(java.lang.String)
	 */
	public ISubject getAssignment(String prosumti) {
		
		if(prosumti.equals("mi"))
			return MI;
		if(prosumti.equals("do"))
			return DO;
		
		if(assignments.containsKey(prosumti)) {
			Stack<JTPSubject> stack = assignments.get(prosumti);
			if(stack.size() > 0)
				return stack.peek();
			else
				throw new RuntimeException("empty assignment stack for " + prosumti);
		}
		throw new RuntimeException("no assignment stack for " + prosumti);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#getQuoteSubject(lojban.analyzer.IFact[])
	 */
	public ISubject getQuoteSubject(IFact facts) {
		return new JTPLiteral("quote", true);
	}

	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#getQuerySubject()
	 */
	public ISubject getQuerySubject() {
		return new JTPLiteral("ask" + queryCount++, false);
	}
	
	////////////////////// LOGIC PART ////////////////////////
	
	public JTPResolver(){
		//Set up reasoners, same as before...
		context = new BasicReasoningContext();
		context.defaultSetup();
		context.setMaxDepth(10);
		
		//A KB for FOL clauses
		ckb = new ClauseOrientationKB();
		context.add(ckb);  //Add to context
		
		//A reasoner for processing complex queries
		AskingQueryProcessor aqp = new AskingQueryProcessor();
		context.add(aqp);  //Add to context

		//Add basic reasoners to asking dispatcher
		DispatcherUtils.addToDispatcher(ckb.getAskingReasoner(),
						context.getAskingDispatcher());
		DispatcherUtils.addToDispatcher(aqp, context.getAskingDispatcher());

		//Add a reasoner that determines if one number is less than
		//another to asking dispatcher
		DispatcherUtils.addToDispatcher(new Less(), context.getAskingDispatcher());
		
		//Add basic reasoners to telling dispatcher
		DispatcherUtils.addToDispatcher(ckb.getTellingReasoner(),
						context.getTellingDispatcher());
		
		// this should no longer be necessary with ?dontcareX in the queries
		//tell("(=> (fact ?r ?y1 ?y2 ?y3 ?y4 ?y5) (fact ?r ?y null null null null))");
		/*for(int i=1;i<32;i++){
			tell("(=> (and (not(= ?x1 ?x2 ?x3 ?x4 ?x5 \"zohe0\")) (fact ?r ?x1 ?x2 ?x3 ?x4 ?x5)) (fact ?r" + 
					((i & 1) ==0 ? " ?x1" : " \"zohe0\"")  +
					((i & 2) ==0 ? " ?x2" : " \"zohe0\"")  +
					((i & 4) ==0 ? " ?x3" : " \"zohe0\"")  +
					((i & 8) ==0 ? " ?x4" : " \"zohe0\"")  +
					((i & 16) ==0 ? " ?x5" : " \"zohe0\"") +
					"))");
		}*/
		tell("(=> (fact ?r ?x1 ?x2 ?x3 ?x4 ?x5) (and (relates ?r ?x1 x1) (relates ?r ?x2 x2) (relates ?r ?x3 x3) (relates ?r ?x4 x4) (relates ?r ?x5 x5)))");
		tell("(=> (factnot ?r ?x1 ?x2 ?x3 ?x4 ?x5) (not (fact ?r ?x1 ?x2 ?x3 ?x4 ?x5)) )");
		tell("(=> (factnot ?r ?x1 ?x2 ?x3 ?x4 ?x5) (and (relatesnot ?r ?x1 x1) (relatesnot ?r ?x2 x2) (relatesnot ?r ?x3 x3) (relatesnot ?r ?x4 x4) (relatesnot ?r ?x5 x5)))");
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.IKnowledgeBase#tell(lojban.analyzer.IFact)
	 */
	public void tell(IFact fact) throws KnowledgeBaseException {
		try {
			context.tellString(((JTPFact) fact).toQuery());
		} catch (ReasoningException e) {
			throw new KnowledgeBaseException();
		}
	}

	/**Tell a fact to the knowledgebase
	 * @param fact the jtp fact string to tell the kb
	 * @throws KnowledgeBaseException
	 */
	public void tell(String fact) throws KnowledgeBaseException {
		try {
			context.tellString(fact);
		} catch (ReasoningException e) {
			throw new KnowledgeBaseException();
		}
	}
	
	/**Ask
	 * @param fact the query to ask (in JTP language)
	 * @return a set of maps mapping a query subject (type JTPLiteral) to all the values it can have in that answer. Each map represents a complete answer.
	 * @throws KnowledgeBaseException
	 */
	public Set<Map<ISubject,ISubject[]>> ask(String fact, boolean multiple) throws KnowledgeBaseException {
		try{
			Set<Map<ISubject,ISubject[]>> ret = new HashSet<Map<ISubject,ISubject[]>>();
			String query=fact;
			
			ReasoningStepIterator rsi = context.getAskingReasoner().process(query);
			ReasoningStep rs = null;
			while ((rs = rsi.next()) != null)
			{
				Map<ISubject,ISubject[]> answer = new HashMap<ISubject,ISubject[]>();
				
			    // If the query is not a literal, then it is transformed into one. The top level
			    // reasoning step then will be the transformation step. Its sole child will have
			    // a goal which is a literal and which is equivalent to the original goal. We
			    // extract the variables from this child's goal.
			    if (query.equals(rs.getGoal()) && rs.getSubProofs().size() == 1)
				{
				    ReasoningStep subRs = (ReasoningStep)rs.getSubProofs().get(0);
				    if (subRs.getGoal() instanceof Literal) 
					rs = subRs; //We are considering the sole child instead of the top-level 
				                    //step (see above)
				}
			    
			    Literal answerLit = (Literal)rs.getGoal();//The answer literal with all the query variables
			    Map bindings = jtp.rs.RSUtils.getRecursiveBindings(rs, null);//Variable bindings
			    for (Iterator it = answerLit.getArgs().iterator(); it.hasNext();)
				{ //Find the right query variable and extract its value
				    Variable v = (Variable)it.next();
				    
				    String variableName = v.toString();
				    JTPLiteral variableSubject;
				    if(variableName.startsWith("?")){
				    	variableName=variableName.substring(1);
				    	variableSubject = new JTPLiteral(variableName,false);
				    }else{
				    	variableSubject = new JTPLiteral(variableName,true);
				    }
				    
				    Set<ISubject> variableSolutions = new HashSet<ISubject>();
				    
				    if(answer.containsKey(variableSubject)){
				    	for(ISubject l : answer.get(variableSubject)){
					    	variableSolutions.add(l);
				    	}
				    	answer.remove(variableSubject);
				    }
					variableSolutions.add(new JTPLiteral(bindings.get(v).toString()));
				    answer.put(variableSubject,variableSolutions.toArray(new ISubject[]{}));
				}
			    
			    ret.add(answer);
			    if(!multiple)
			    	break;
			}
			return ret;
		}catch(ReasoningException e){
			throw new KnowledgeBaseException();
		}
	}
	
	/**Return all facts that match the given mask (if any parameter is null, it's treated as &quot;do not care&quot;
	 * @param r
	 * @param truth
	 * @param x1
	 * @param x2
	 * @param x3
	 * @param x4
	 * @param x5
	 * @return all facts that match the given mask
	 * @throws KnowledgeBaseException
	 */
	public Set<IFact> getFacts(IRelation r,boolean truth,ISubject x1, ISubject x2, ISubject x3, ISubject x4, ISubject x5) throws KnowledgeBaseException {
		StringBuilder queryBuilder=new StringBuilder();
		if(! truth)
			queryBuilder.append("(factnot ");
		else
			queryBuilder.append("(fact ");
		if(r == null)
			queryBuilder.append("?r");
		else
			queryBuilder.append(r.toString());
		queryBuilder.append(" ");
		
		if(x1 == null)
			queryBuilder.append("?x1");
		else
			queryBuilder.append(x1.toString());
		queryBuilder.append(" ");
		
		if(x2 == null)
			queryBuilder.append("?x2");
		else
			queryBuilder.append(x2.toString());
		queryBuilder.append(" ");
		
		if(x3 == null)
			queryBuilder.append("?x3");
		else
			queryBuilder.append(x3.toString());
		queryBuilder.append(" ");
		
		if(x4 == null)
			queryBuilder.append("?x4");
		else
			queryBuilder.append(x4.toString());
		queryBuilder.append(" ");
		
		if(x5 == null)
			queryBuilder.append("?x5");
		else
			queryBuilder.append(x5.toString());
		queryBuilder.append(")");
		
		//if(! truth)
		//	queryBuilder.append(")");
		
		String query=queryBuilder.toString();
		
		try{
			Set<IFact> ret=new HashSet<IFact>();
			
			ReasoningStepIterator rsi = context.getAskingReasoner().process(query);
			ReasoningStep rs = null;
			while ((rs = rsi.next()) != null)
			{
			    // If the query is not a literal, then it is transformed into one. The top level
			    // reasoning step then will be the transformation step. Its sole child will have
			    // a goal which is a literal and which is equivalent to the original goal. We
			    // extract the variables from this child's goal.
			    if (query.equals(rs.getGoal()) && rs.getSubProofs().size() == 1)
				{
				    ReasoningStep subRs = (ReasoningStep)rs.getSubProofs().get(0);
				    if (subRs.getGoal() instanceof Literal) 
					rs = subRs; //We are considering the sole child instead of the top-level 
				                    //step (see above)
				}
			    
			    Literal answerLit = (Literal)rs.getGoal();//The answer literal with all the query variables
			    Map bindings = jtp.rs.RSUtils.getRecursiveBindings(rs, null);//Variable bindings
			    for (Iterator it = answerLit.getArgs().iterator(); it.hasNext();)
				{ //Find the right query variable and extract its value
				    Variable v = (Variable)it.next();
				    if ("r".equals(v.getName()))
					{
				    	r=getRelationByName(bindings.get(v).toString());
					}
				    if ("x1".equals(v.getName()))
					{
					    if(bindings.get(v).toString().equals("zohe0"))
					    	x1 = null;
					    else
					    	x1=new JTPLiteral(bindings.get(v).toString());
					}
				    if ("x2".equals(v.getName()))
					{
					    if(bindings.get(v).toString().equals("zohe0"))
					    	x2 = null;
					    else
					    	x2=new JTPLiteral(bindings.get(v).toString());
					}
				    if ("x3".equals(v.getName()))
					{
					    if(bindings.get(v).toString().equals("zohe0"))
					    	x3 = null;
					    else
					    	x3=new JTPLiteral(bindings.get(v).toString());
					}
				    if ("x4".equals(v.getName()))
					{
					    if(bindings.get(v).toString().equals("zohe0"))
					    	x4 = null;
					    else
					    	x4=new JTPLiteral(bindings.get(v).toString());
					}
				    if ("x5".equals(v.getName()))
					{
					    if(bindings.get(v).toString().equals("zohe0"))
					    	x5 = null;
					    else
					    	x5=new JTPLiteral(bindings.get(v).toString());
					}
				}
			    ret.add(createFact(truth,(JTPRelation) r,(JTPSubject) x1,(JTPSubject) x2,(JTPSubject) x3, (JTPSubject)x4,(JTPSubject) x5));
			}
			return ret;
		}catch(ReasoningException e){
			throw new KnowledgeBaseException();
		}
	}
	
	/**Return all facts that have the specified subject in them
	 * @param s the subject to search for
	 * @return all facts that s is a part of
	 * @throws KnowledgeBaseException
	 */
	public Set<IFact> getFacts(ISubject s,boolean truth) throws KnowledgeBaseException {
		try{
			Set<IFact> ret=new HashSet<IFact>();
			
			ArrayList<IRelation> relations = new ArrayList<IRelation>();
			ArrayList<Integer> positions = new ArrayList<Integer>();
			
			String query;
			if(truth)
				query="(relates ?x " + s.toString() + " ?y)";
			else
				query="(relatesnot ?x " + s.toString() + " ?y)";
			
			ReasoningStepIterator rsi = context.getAskingReasoner().process(query);
			ReasoningStep rs = null;
			while ((rs = rsi.next()) != null)
			{
			    // If the query is not a literal, then it is transformed into one. The top level
			    // reasoning step then will be the transformation step. Its sole child will have
			    // a goal which is a literal and which is equivalent to the original goal. We
			    // extract the variables from this child's goal.
			    if (query.equals(rs.getGoal()) && rs.getSubProofs().size() == 1)
				{
				    ReasoningStep subRs = (ReasoningStep)rs.getSubProofs().get(0);
				    if (subRs.getGoal() instanceof Literal) 
					rs = subRs; //We are considering the sole child instead of the top-level 
				                    //step (see above)
				}
			    
			    IRelation relation = null; //the relation that s is involved in
			    int position=0; //the position of s in the fact with relation relation and subject s
			    
			    Literal answerLit = (Literal)rs.getGoal();//The answer literal with all the query variables
			    Map bindings = jtp.rs.RSUtils.getRecursiveBindings(rs, null);//Variable bindings
			    for (Iterator it = answerLit.getArgs().iterator(); it.hasNext();)
				{ //Find the right query variable and extract its value
				    Variable v = (Variable)it.next();
				    if ("x".equals(v.getName()))
					{
				    	relation=getRelationByName(bindings.get(v).toString());
					}
				    if ("y".equals(v.getName()))
					{
					    position=Integer.parseInt((bindings.get(v)).toString().substring(1));
					}
				}
			    relations.add(relation);
			    positions.add(position);
			}
		    
			for(int i=0;i<positions.size();i++){
				IRelation relation;
				int position;
				
				position=positions.get(i);
				relation=relations.get(i);
				
				Set<IFact> ifacts=null;
			    switch(position){
			    case 1:
			    	ifacts=getFacts(relation,truth,(JTPSubject) s,null,null,null,null);
			    	break;
			    case 2:
			    	ifacts=getFacts(relation,truth,null,(JTPSubject) s,null,null,null);
			    	break;
			    case 3:
			    	ifacts=getFacts(relation,truth,null,null,(JTPSubject) s,null,null);
			    	break;
			    case 4:
			    	ifacts=getFacts(relation,truth,null,null,null,(JTPSubject) s,null);
			    	break;
			    case 5:
			    	ifacts=getFacts(relation,truth,null,null,null,null,(JTPSubject) s);
			    	break;
			    }
			    ret.addAll(ifacts);
			}
			
			return ret;
		}catch(ReasoningException e){
			throw new KnowledgeBaseException();
		}
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.IKnowledgeBase#ask(lojban.analyzer.IFact)
	 */
	public Set<Map<ISubject,ISubject[]>> ask(IFact fact, boolean multiple) throws KnowledgeBaseException {
		return ask(((JTPFact) fact).toQuery(), multiple);
	}
	
	/* (non-Javadoc)
	 * @see lojban.analyzer.IResolver#describe(lojban.analyzer.ISubject)
	 */
	public ISubject describe(ISubject subject) throws KnowledgeBaseException{
		return describe(subject,new HashSet<ISubject>(),new HashSet<IFact>());
	}
	
	/**Flatten a composite fact into an set of facts
	 * @param fact
	 * @return
	 */
	Set<IFact> flattenFact(IFact fact){
		Set<IFact> ret = new HashSet<IFact>();
		if(fact.isComposite()){
			ret.addAll(flattenFact(fact.getLeft()));
			ret.addAll(flattenFact(fact.getRight()));
			return ret;
		}else{
			ret.add(fact);
			return(ret);
		}
	}
	
	/**Describe a subject
	 * @param subject the subject to describe
	 * @param question the question that is to be answered (specified so the description won't repeat the question)
	 * @return
	 * @throws KnowledgeBaseException
	 */
	public ISubject describe(ISubject subject,IFact question) throws KnowledgeBaseException{
		Set<IFact> facts=new HashSet<IFact>();
		for(IFact fact : flattenFact(question)){
			for(int i=0;i<5;i++){
				if(fact.getSubject(i)!=null){
					if(fact.getSubject(i).toString().startsWith("?ask"))
						fact = ((JTPFact)fact).substitute((JTPSubject)fact.getSubject(i),(JTPSubject)subject);
					if(fact.getSubject(i).toString().startsWith("?q"))
						fact = ((JTPFact)fact).substitute((JTPSubject)fact.getSubject(i),new JTPLiteral("x" + i,false));
				}
			}
			facts.add(fact);
		}
		return describe(subject,new HashSet<ISubject>(),facts);
	}
	
	/**Describe a subject
	 * @param subject the subject to describe
	 * @param alreadyDescribedSubjects the subjects that have already been described. Prevents infinite recursion.
	 * @param alreadyDescribedFacts that have already been used in a description. Prevents stupid repetitions leading to useless and/or ambiguous answers.
	 * @return a fact that describes the subject
	 * @throws KnowledgeBaseException
	 */
	public ISubject describe(ISubject subject,Set<ISubject> alreadyDescribedSubjects, Set<IFact> alreadyUsedFacts) throws KnowledgeBaseException {
		/* Find all relationships that the subject is in
		 * replace the subject with a query, and then ask that
		 */
		alreadyDescribedSubjects.add(subject);
		
		
		Set<IFact> uniqueFactsSet = new HashSet<IFact>();
		
		Set<IFact> ambiguousFacts = new HashSet<IFact>();
		ambiguousFacts.addAll(getFacts(subject,true));
		ambiguousFacts.addAll(getFacts(subject,false));
		ambiguousFacts.removeAll(alreadyUsedFacts);
		
		//test if each fact is unique - put unique facts are in uniqueFacts, and the non-unique facts are in ambiguousFacts
		Iterator i=ambiguousFacts.iterator();
		while(i.hasNext()){
			IFact fact = (IFact) i.next();
			Set<Map<ISubject, ISubject[]>> answers=ask(((JTPFact) fact).toQuery().replace(subject.toString(),"?ask0"), true);
			if(isUniqueSolution(answers,new JTPLiteral("ask0",false))){
				//this fact is unique and has not been used
				uniqueFactsSet.add(fact);
				i.remove();
			}
		}
		
		List<IFact> uniqueFactsList = new ArrayList<IFact>(uniqueFactsSet);
		
		sortIFactList(uniqueFactsList, subject);
		
		for (IFact retFact : uniqueFactsList){
			alreadyUsedFacts.add(retFact);
			for(int j=0;j<5;j++){
				if(retFact.getSubject(j) != null && retFact.getSubject(j).isIdentifier() && !retFact.getSubject(j).isLiteral() && ! alreadyDescribedSubjects.contains(retFact.getSubject(j))){
					ISubject newSubject=describe(retFact.getSubject(j),alreadyDescribedSubjects,alreadyUsedFacts);
					retFact = ((JTPFact)retFact).substitute((JTPSubject)retFact.getSubject(j),(JTPSubject)newSubject);
				}
			}
			return new JTPConcreteSubject((JTPFact) retFact, (JTPLiteral)subject);
		}
		throw new KnowledgeBaseException("Do not know how to describe this subject: " + subject);
	}
	
	/**Determine if the question literal maps to exactly one subject
	 * @param answers the answers returned by an ask call
	 * @param question the question subject
	 * @return true if exactly one solution exists for this question subject, false otherwise
	 */
	boolean isUniqueSolution(Set<Map<ISubject, ISubject[]>> answers, JTPLiteral question){
		Set<ISubject> discoveredValues=new HashSet<ISubject>();
		for(Map<ISubject, ISubject[]> answer : answers){
			if(answer.containsKey(question))
				for(ISubject s : answer.get(question)){
					discoveredValues.add(s);
				}
		}
		return(discoveredValues.size()==1);
	}
	
	/**Sort a List of IFact by the number of subjects in the IFact, and the truth value of the IFact
	 * @param facts
	 */
	void sortIFactList(List<IFact> facts, final ISubject s){
//		sort the facts by the number of subjects in them
		Collections.sort(facts,new Comparator<IFact>(){
			/**Count the subjects in a fact for purposes of sorting a list of facts
			 * @param fact
			 * @return
			 */
			private int countSubjects(IFact fact){
				int retval=0;
				for(int i=0;i<5;i++)
					if(fact.getSubject(i)!=null){
						if(fact.getSubject(i).isIdentifier())
							//weight identifiers more heavily, they are less useful in a description (cause recursion and longer answers)
							retval+=6;
						else if(fact.getSubject(i).isLiteral())
							//weight literals less heavily, as they are more useful, and give nice concise answers
							retval+=1;
					}
				return retval;
				
			}
			
			/**Compare two subjects, as per the "Comparator" interface
			 * @param o1
			 * @param o2
			 * @return
			 */
			public int compare(IFact o1, IFact o2){
				int c1=countSubjects(o1)+(o1.isTrue()?10:-10)+(o1.getSubject(0).equals(s)?10:0);
				int c2=countSubjects(o2)+(o2.isTrue()?10:-10)+(o2.getSubject(0).equals(s)?10:0);
				if(c1==c2)
					return 0;
				else if(c1<c2)
					return -1;
				else
					return 1;
			}
		}
		);
	}

	public IRelation getQueryRelation() {
		return new JTPRelation("?askrel" + queryCount);
	}

	public ISubject getProsumtiSubject(String string) {
		return JTPLiteral.getProsumti(string);
	}
}
