/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.parser;

import java.io.BufferedReader;
import java.io.FileReader;

public abstract class AbstractDictionary {
	
	/**
	 * @return dictionary filename
	 */
	public abstract String getDictionaryFile();
	
	/**
	 * @return separator list for heuristics
	 */
	public abstract String[] getSeparators();
	
	/**
	 * Get a word to represent a lojban concept.
	 * @param s a lojban word
	 * @return an approximate translation
	 */
	public abstract String getWord(String s);
	
	/**
	 * @return index of noun column start
	 */
	public abstract int getNounStart();
	
	/**
	 * @return index of noun column end
	 */
	public abstract int getNounEnd();
	
	/**
	 * @return index of entry column start
	 */
	public abstract int getEntryStart();
	
	/**
	 * @return index of entry column end (or -1 if the end of the line)
	 */
	public abstract int getEntryEnd();
	
	/**
	 * @return number of lines to skip at start of dictionary file
	 */
	public abstract int getSkip();
	
	/**
	 * Get a line in the dictionary file.
	 * @param gismu gismu to search for
	 * @return the dictionary line
	 */
	private String getEntry(String gismu) {
		try {
			String line = null;
			BufferedReader in = new BufferedReader(new FileReader(getDictionaryFile()));
			for(int i = 0; i < getSkip(); i++)
				in.readLine();	// skip header
			while((line = in.readLine()) != null) {
				if(line.substring(1, 6).equals(gismu)) {
					in.close();
					return line;
				}
			}
			in.close();
			return null;
		} catch(Exception e) {
			return null;
		}
	}
	
	/**
	 * Trim/clean up a string.
	 * @param s string 
	 * @return cleaned up string
	 */
	private String clean(String s) {
		if(s.startsWith(","))
			s = s.substring(1).trim();
		if(s.indexOf("(") >= 0)
			return (s.substring(0, s.indexOf("(")).trim() + " " + s.substring(s.indexOf(")")+1).trim()).trim();
		else
			return s.trim();
	}
	
	/**
	 * Get an approximate noun for a gismu.
	 * @param gismu the gismu
	 * @return an approximage noun translation
	 */
	public String getNoun(String gismu) {
		gismu = gismu.trim();
		String result = "[noun:" + gismu + "]";
		String entry = getEntry(gismu);
		if(entry != null)
			result = entry.substring(getNounStart(), getNounEnd()).trim();
		return clean(result);
	}
	
	/**
	 * Try to split the x1..x2 clause into a verb and a tag
	 * @param entry the dictionary entry line charcters between x1..x2 (exclusive)
	 * @return a best guess at where to partition then
	 */
	private int heuristicSeparation(String entry) {
		// look for separation words		
		// find where any appear
		String sep[] = getSeparators();
		int pos[] = new int[sep.length];
		for(int i = 0; i < sep.length; i++)
			pos[i] = entry.lastIndexOf(sep[i]) + sep[i].length();
		
		// find which appears last
		int max = 0;
		for(int i = 1; i < pos.length; i++)
			if(pos[i] > pos[max])
				max = i;
		
		return pos[max] - sep[max].length();
	}
	
	/**
	 * Get an approximate verb phrase for a gismu
	 * @param gismu the gismu
	 * @return an approximate verm
	 */
	public String getVerb(String gismu) {
		gismu = gismu.trim();
		String result = "[verb:" + gismu + "]";
		String entry = getEntry(gismu);
		if(entry != null) {
			if(getEntryEnd() == -1)
				entry = entry.substring(getEntryStart());
			else
				entry = entry.substring(getEntryStart(), getEntryEnd());
			if(entry.indexOf(";") > 0)
				entry = entry.substring(0, entry.indexOf(";"));
			if(entry.indexOf("x2") > 0)
				entry = entry.substring(entry.indexOf("x1")+2, entry.indexOf("x2")).trim();
			else
				entry = entry.substring(entry.indexOf("x1")+2);
			

			int max = heuristicSeparation(entry);
			if(max > 0)
				result = entry.substring(0, max).trim();
			else
				result = entry.trim();	
		}
		return clean(result);
	}
	
	/**
	 * Get the tag to prepend to a subject.
	 * @param gismu the gismu
	 * @param i the subject (2-5)
	 * @return the tag for the given subject
	 */
	public String getSubjectTag(String gismu, int i) {
		if(gismu == null)
			gismu = "--mo--";
		gismu = gismu.trim();
		String result = getWord("motag");;
		String entry = getEntry(gismu);
		if(entry != null) {
			if(getEntryEnd() == -1)
				entry = entry.substring(getEntryStart());
			else
				entry = entry.substring(getEntryStart(), getEntryEnd());
			
			if(entry.indexOf(";") > 0)
				entry = entry.substring(0, entry.indexOf(";"));
			
			if(entry.indexOf("x" + i) < 0)
				return getWord("motag");
			
			entry = entry.substring(entry.indexOf("x"+(i-1))+2, entry.indexOf("x" + i)).trim();	
			
			if(i == 2) {
				int max = heuristicSeparation(entry);
				if(max > 0)
					result = entry.substring(max).trim();
				else
					result = "";
			} else {
				result = entry;
			}
			
		}
		return clean(result);
	}	
}
