/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.parser;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.io.Reader;
import java.io.StringReader;
import java.util.StringTokenizer;
import java.util.Vector;

import xtc.parser.PParser;
import xtc.parser.Result;
import xtc.parser.SemanticValue;
import xtc.tree.GNode;
import xtc.tree.Printer;

public class GenericParseNode {

	/**
	 * Construct recursively from a line from the parser.
	 * @param a line of text from the rats! parser.
	 */
	private GenericParseNode(String line) {
		StringTokenizer st = new StringTokenizer(line, " ");
		if(st.hasMoreTokens())
			ConstructNode(st.nextToken(), st);
	}
	
	/**
	 * For recursive construction of tree.
	 * @param first token that caused parent to delegate to us
	 * @param st tokenizer with rest of tokens.
	 */
	private GenericParseNode(String first, StringTokenizer st) {
		ConstructNode(first, st);
	}
	
	/**
	 * Function to perform recursive construction of tree.
	 * @param first token that caused parent to delegate to us 
	 * @param st tokenizer with rest of tokens.
	 */
	private void ConstructNode(String first, StringTokenizer st) {
		
		// drop the =(
		type = first.substring(0, first.length()-2);
		
		// look at remaining tokens
		while(st.hasMoreTokens()) {
					
			// get segment
			String token = st.nextToken();
			
			// closing paren
			if(token.equals(")"))
				return;
			
			// child node
			if(token.endsWith("=(")) {
				if(children == null)
					children = new Vector<GenericParseNode>();
				children.add(new GenericParseNode(token, st));
			} else
				content = token;
			
		}
	}
	
	/**
	 * Debug print this node's tree structure.
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		if(children != null) {
			sb.append(" " + type + "[");
			for(GenericParseNode n : children)
				sb.append(n.toString());
			sb.append("] ");
		} else {
			sb.append(" " + type + "=" + content + " ");
		}
		return sb.toString();
	}
	
	/**
	 * Turn a lojban string into a ParseNode representation.
	 * @param str the lojban string
	 * @return a vector containing nodes representing the resultings structure
	 * @throws Exception
	 */
	public static Vector<GenericParseNode> parse(String str) {
		
		synchronized(GenericParseNode.class) {
				
		try {

			// write string to file for reading
			String result = parseDelegate(str);

			// process input
			BufferedReader read = new BufferedReader(new StringReader(result));
			String line = null;
			Vector<GenericParseNode> nodes = new Vector<GenericParseNode>();
			while((line = read.readLine()) != null) {
				nodes.add(new GenericParseNode(line));
			}
			
			// check for errors
			for(GenericParseNode n : nodes) {
				if(n == null)
					throw new ParseException("parse error");
				if(!letterMatch(n.getText(), str)) {
					throw new ParseException("parse error after: " + n.getText());
				}
			}
			
			return nodes;
		} catch(IOException e) {
			e.printStackTrace();
			return null;
		}
		}
	}
	
	/**
	 * Delegate directly to the parser code.
	 * Wrap as necessary to get intelligible output.
	 * (This wrapping code modelled on the main of xtc.parser.PParser)
	 * @param utterance utterance to delegate
	 */
	private static String parseDelegate(String utterance) {
		PParser.text = false;
		PParser.verbose = true;
		PParser.pretty = false;

		// stream to collect output (instead of System.out)
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		PrintStream localout = new PrintStream(baos);
		
		BufferedReader in = null;
		Reader in2 = null;
		Reader input_reader = null;
		try {

			String input_string;
			in = new BufferedReader(new StringReader(utterance));
			input_string = in.readLine();
			while (input_string != null) {
				input_reader = new BufferedReader(
						new StringReader(input_string));
				PParser p = new PParser(input_reader, "Morphology Pass");
				Result r = p.pmorphology(0);
				String total1 = null, total2;

				if (r.hasValue()) {
					SemanticValue v = (SemanticValue) r;

					if (v.value instanceof GNode) {
						new Printer(localout).incr().p((GNode) v.value)
								.flush();
					} else {
						total1 = PParser.finalMakeString(v.value);
						total1 = total1.replaceAll("Morph=\\(", "=(");
					}

				} else {
					System.out.println("Error in morphology phase.\n");
					return null;
				}

				total1 = total1.replaceFirst("^ text=\\( ", "");
				total1 = total1.replaceFirst(" \\) $", "");
				in2 = new BufferedReader(new StringReader(total1));
				PParser p2 = new PParser(in2, "Main Pass");
				Result r2 = p2.ptext(0);

				if (r2.hasValue()) {
					SemanticValue v = (SemanticValue) r2;

					if (v.value instanceof GNode) {
						new Printer(localout).incr().p((GNode) v.value)
								.flush();
					} else {
						total2 = PParser.finalMakeString(v.value);
						localout.println(total2);
					}

				} else {
					System.out.println("Error in main phase.\n");
					return null;
				}

				input_string = in.readLine();
			}

		} catch (Throwable x) {
			x.printStackTrace();

			while (null != x.getCause()) {
				x = x.getCause();
			}
			System.out.println("  " + x.toString());
			return null;
		} finally {
			try {
				in.close();
				in2.close();
				input_reader.close();
			} catch (Throwable x) {
			}
		}
		
		// extract result from stream
		return baos.toString();
	}
	
	/** ensure the letter characters in a match those in b */
	private static boolean letterMatch(String a, String b) {
		a = a.trim();
		b = b.trim();
		
		StringBuilder a1 = new StringBuilder(), b1 = new StringBuilder();
		for(int i = 0; i < a.length(); i++)
			if(Character.isLetter(a.charAt(i)))
				a1.append(Character.toLowerCase(a.charAt(i)));
		for(int i = 0; i < b.length(); i++)
				if(Character.isLetter(b.charAt(i)))
					b1.append(Character.toLowerCase(b.charAt(i)));
		return a1.toString().equals(b1.toString());
	}
	
	
	/**
	 * Get the type/language class represented by this node.
	 * @return type/class
	 */
	public String getType() {
		return type;
	}

	/**
	 * Get the content represented by this node.
	 * @return content of node
	 */
	public String getContent() {
		return content;
	}

	/**
	 * Get children this node has
	 * @return array of children
	 */
	public GenericParseNode[] getChildren() {
		if(children == null)
			return new GenericParseNode[]{};
		return children.toArray(new GenericParseNode[]{});
	}
	
	
	/**
	 * Get the text contained by this node. 
	 */
	public String getText() {
		if(children == null) {
			
			if(content == null || content.length() == 0)
				return "";
			if("aeiouy".indexOf(content.charAt(0)) >= 0)
				return "." + content;
			return content;
			
		} else {
			String s = null;
			for(GenericParseNode n : children) {
				if(s == null)
					s = n.getText();
				else
					s += " " + n.getText();
			}
			return s;
		}
	}
	
	private String type = null;
	private String content = null;
	private Vector<GenericParseNode> children = null;
}
