/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package lojban.synthesis;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import lojban.analyzer.IFact;
import lojban.analyzer.IRelation;
import lojban.analyzer.IResolver;
import lojban.analyzer.ISubject;
import lojban.analyzer.IFact.Connective;
import lojban.kb.jtp.JTPConcreteSubject;
import lojban.kb.jtp.JTPFact;
import lojban.kb.jtp.JTPLiteral;
import lojban.kb.jtp.JTPResolver;
import lojban.kb.jtp.JTPSubject;

/**
 * @author mike
 * Interface class to build lojban text from facts and subjects.
 */
public class Synthesis {

	/**
	 * Turn a restricting fact set into lojban text. 
	 * @param f1 the restrictngs facts
	 * @param q the subject the refer to
	 * @param resolver the resolver providing the subjects
	 * @return a lojban string describing these facts
	 */
	public String synthesizeRestrictingFact(IFact f1[], ISubject q, IResolver resolver) {
		StringBuilder sb = new StringBuilder();
		boolean first = true;
		for(IFact f : f1) {
			
			if(first) {
				sb.append("poi ");
				first = false;
			} else
				sb.append("zi'e poi ");
			if(f.isComposite()) {
				sb.append(synthesizeCompositeRestrictingFact(f, q, resolver));
			} else {
				sb.append(synthesizeSimpleRestrictingFact(f, q, resolver));
			}
		}
		return sb.toString();
	}
	
	/**
	 * Turn a simple restricting fact into lojban text.
	 * @param f fact to synthesize
	 * @param q subject it restricts
	 * @param resolver resolver producing subject
	 * @return lojban string corresponding to fact
	 */
	public String synthesizeSimpleRestrictingFact(IFact f, ISubject q, IResolver resolver) {
		StringBuilder sb = new StringBuilder();
		if(f.getSubject(0) != null)
			addRestrictingSubject(f, q, 0, sb, resolver);
		sb.append("cu " + f.getRelation().toString() + " ");
		if(f.getSubject(1) != null || f.getSubject(3) != null || f.getSubject(2) != null || f.getSubject(1) != null)
			addRestrictingSubject(f, q, 1, sb, resolver);
		if(f.getSubject(2) != null || f.getSubject(3) != null || f.getSubject(2) != null)
			addRestrictingSubject(f, q, 2, sb, resolver);
		if(f.getSubject(3) != null || f.getSubject(3) != null)
			addRestrictingSubject(f, q, 3, sb, resolver);
		if(f.getSubject(4) != null)
			addRestrictingSubject(f, q, 4, sb, resolver);
		return sb.toString();
	}
	
	/**
	 * Turn a compound restricting fact into lojban text.
	 * @param f fact to synthesize
	 * @param q subject it restricts
	 * @param resolver resolver producing subject
	 * @return lojban string corresponding to fact
	 */
	public String synthesizeCompositeRestrictingFact(IFact f, ISubject q, IResolver resolver) {
		StringBuffer sb = new StringBuffer();
		sb.append("g");
		switch(f.getConnective()) {
			case AND: sb.append("e"); break;
			case OR: sb.append("a"); break;
			case IFF: sb.append("o"); break;
			case WON: sb.append("u"); break;
		}
		if(f.getNegLeft())
			sb.append("nai");
		
		if(f.getLeft().isComposite())
			sb.append(" " + synthesizeCompositeRestrictingFact(f.getLeft(), q, resolver));
		else
			sb.append(" " + synthesizeSimpleRestrictingFact(f.getLeft(), q, resolver));
		
		sb.append("gi");
		if(f.getNegLeft())
			sb.append("nai");
		
		if(f.getRight().isComposite())
			sb.append(" " + synthesizeCompositeRestrictingFact(f.getRight(), q, resolver) + " ");
		else
			sb.append(" " + synthesizeSimpleRestrictingFact(f.getLeft(), q, resolver) + " ");
		
		return sb.toString();
	}
	
	/**
	 * Flatten a tree of ands into a list of the and-ed facts.
	 * @param f root fact
	 * @return list of anded facts
	 */
	private LinkedList<IFact> flattenAnds(IFact f) {
		LinkedList<IFact> flat = new LinkedList<IFact>();
		if(f.isComposite() && f.getConnective() == Connective.AND && !f.getNegLeft() && !f.getNegRight()) {
			flat.addAll(flattenAnds(f.getLeft()));
			flat.addAll(flattenAnds(f.getRight()));
		} else {
			flat.add(f);
		}
		return flat;
	}
	
	/**
	 * Add a subject to a string by its position.
	 * @param f fact containing subject
	 * @param q subject being restricted
	 * @param i subject position
	 * @param sb buffer to receive text for subject
	 * @param resolver resolver which created facts
	 */
	private void addRestrictingSubject(IFact f, ISubject q, int i, StringBuilder sb, IResolver resolver) {
		if(f.getSubject(i) == null)
			sb.append("zo'e ");
		else if(f.getSubject(i) == q)
			sb.append("ke'a ");
		else if(f.getSubject(i).isProsumti())
			sb.append(f.getSubject(i).toString().replace("\"", ""));
		else if(f.getSubject(i).isLiteral())
			sb.append("zo " + makeLojbanCmene(f.getSubject(i).toString()) + " ");
		else
			sb.append(synthesizeSubject(f.getSubject(i), resolver));
	}
	
	/**
	 * Ensure cmene is properly lojbanized (with initial full stop before vowel, final full stop after consonant)
	 * @param str string to lojbanize
	 * @return lojbanized string
	 */
	private String makeLojbanCmene(String str) {
		String s = str.replace("\"", "");
		if("aeiou".indexOf(s.charAt(0)) >= 0)
			s = "." + s;
		if("bcdfgjklmnprstvxz".indexOf(s.charAt(s.length()-1)) >= 0)
			s = s + ".";
		return s;
	}
	
	/**
	 * Turn facts/subject into a description.
	 * @param s subject to describe
	 * @return lojban string describing s with f
	 */
	public String synthesizeSubject(ISubject s, IResolver resolver) {
		if(s.isLiteral())
			return s.toString();
		if(s.isRestricted()) {
			
			// collapse ands to an array
			LinkedList<IFact> facts = flattenAnds(s.getRestricting());
			
			// see if there's a cmene we can use
			Iterator<IFact> f1 = facts.iterator();
			while(f1.hasNext()) {
				IFact f = f1.next();
				if(f.getRelation().toString().equals("cmene") && f.getSubject(1) == s.getRestrictingIdentifier() && f.getSubject(0) != null) {
					f1.remove();
					return "la " + makeLojbanCmene(f.getSubject(0).toString()) + " " + synthesizeRestrictingFact(facts.toArray(new IFact[]{}), s.getRestrictingIdentifier(), resolver);
				}
			}
			
			// see if there's a something we can use to make a nice, simple non-veridical
			f1 = facts.iterator();
			while(f1.hasNext()) {
				IFact f = f1.next();
				if(f.getSubject(0) == s.getRestrictingIdentifier() && f.getSubject(1) == null && f.getSubject(2) == null && f.getSubject(3) == null && f.getSubject(4) == null) {
					f1.remove();		// don't repeat fact
					return "le " + f.getRelation().toString() + " " + synthesizeRestrictingFact(facts.toArray(new IFact[]{}), s.getRestrictingIdentifier(), resolver);
				}
			}
			
			// see if there's a something we can use to make a less simple non-veridical
			f1 = facts.iterator();
			while(f1.hasNext()) {
				IFact f = f1.next();
				if(f.getSubject(0) == s.getRestrictingIdentifier())
					return "le " + f.getRelation().toString() + " " + synthesizeRestrictingFact(facts.toArray(new IFact[]{}), s.getRestrictingIdentifier(), resolver);
			}
			
			// revert to overly verbose, hard to understand stuff
			return "le sidbo " + synthesizeRestrictingFact(facts.toArray(new IFact[]{}), s.getRestrictingIdentifier(), resolver);
			
		}
		else throw new SynthesisException("insufficient info to describe " + s.toString());
	}
	
	
	
	/**
	 * Some tests.
	 */
	public static void main(String args[]) {
		// synthesis tests
		
		JTPResolver resolver = new JTPResolver();		
		
		// create guen
		JTPLiteral idg = new JTPLiteral("s1");
		JTPFact nameg = (JTPFact) resolver.createFact(
				true,
				resolver.getRelationByName("cmene"),
				resolver.getLiteralSubject("guen"),
				idg,
				null, null, null);
		JTPSubject guen = new JTPConcreteSubject(nameg, idg);
		
		// create "named arthur and loves guen"
		JTPLiteral ida = new JTPLiteral("s0");
		JTPFact namea = (JTPFact) resolver.createFact(			// named arthur
				true,
				resolver.getRelationByName("cmene"),
				resolver.getLiteralSubject("artr"),
				ida,
				null, null, null);		
		JTPFact love = (JTPFact) resolver.createFact(			// loves guen
				true,
				resolver.getRelationByName("prami"),
				ida,
				guen,
				null, null, null);
		JTPFact f1 = (JTPFact) resolver.createConnectiveFact(Connective.AND, false, false, namea, love);
		JTPSubject artr = new JTPConcreteSubject(f1, ida);
		
		Synthesis s = new Synthesis();
		
		System.out.println(s.synthesizeSubject(artr, resolver));
	}

	/**
	 * Collect all queries from a question, drop constants, and substitute answers for variables.
	 * @param question question to answer
	 * @param answers descriptions for answers
	 * @param selbri 
	 * @param resolver resolver responsible for this operation
	 * @return fact that answers the question
	 */
	public IFact createAnswerFact(IFact question, Map<ISubject, ISubject> answers, HashMap<IRelation, IRelation> selbri, IResolver resolver) {
		
		// find the questions
		LinkedList<IFact> queryFacts = new LinkedList<IFact>();
		collectSimpleQueries(question, queryFacts);
		
		// create the answers
		LinkedList<IFact> answerFacts = new LinkedList<IFact>();
		for(IFact query : queryFacts) {
			
			// find important subjects
			ISubject a[] = new ISubject[5];
			for(int i = 0; i < 5; i++) {
				if(query.getSubject(i) != null && answers.containsKey(query.getSubject(i)))
					a[i] = answers.get(query.getSubject(i));
			}
			
			// todo get answers somehow!
			IRelation relation = null;
			if(query.getRelation().toString().startsWith("?"))
				relation = selbri.get(resolver.getRelationByName(query.getRelation().toString()));
			else
				relation = query.getRelation();
			
			// assemble into fact and save
			answerFacts.add(resolver.createFact(true, relation, a[0], a[1], a[2], a[3], a[4]));
		}
		
		// join to single composite fact
		IFact result = null;
		for(IFact answerFact : answerFacts) {
			if(result == null)
				result = answerFact;
			else
				result = resolver.createConnectiveFact(IFact.Connective.AND, false, false, result, answerFact);
		}
		
		return result;
	}
	
	/**
	 * Add all simple (non-composite/connective) queries in fact to queries.
	 * @param fact root fact
	 * @param queries receives all simple query facts
	 */
	protected void collectSimpleQueries(IFact fact, LinkedList<IFact> queries) {		
		if(fact.isComposite()) {
			if(fact.getLeft().isQuery())
				collectSimpleQueries(fact.getLeft(), queries);
			if(fact.getRight().isQuery())
				collectSimpleQueries(fact.getRight(), queries);
		} else {
			if(fact.isQuery())
				queries.add(fact);
		}
	}
	
	/**
	 * Build a lojban string for a fact.
	 * @param fact the fact
	 * @param resolver the resolver
	 * @return lojban string for fact
	 */
	public String synthesizeFact(IFact fact, IResolver resolver) {
		
		// first flatten out ands for cosmetic reasons
		LinkedList<IFact> facts = flattenAnds(fact);
		
		// handle remaining composites
		StringBuilder sb = new StringBuilder();
		for(IFact f : facts) {
			sb.append(".i ");
			appendFact(f, sb, resolver);
			sb.append(" ");
		}
		
		// return assembled string
		return sb.toString().trim();
	}
	
	/**
	 * Add a subject to a string by its position.
	 * @param f fact containing subject
	 * @param i subject position
	 * @param sb buffer to receive text for subject
	 */
	protected void appendSubject(IFact f, int i, StringBuilder sb, IResolver resolver) {
		if(f.getSubject(i) == null) {
			if(i > 0 && i < 5)
				for(int j = i+1; j < 5; j++)			// append only if there's something after that we need to keep the space for			
					if(f.getSubject(j) != null) {
						sb.append("zo'e ");
						break;
					}
		} else if(f.getSubject(i).isProsumti())
			sb.append(f.getSubject(i).toString().replace("\"", ""));
		else if(f.getSubject(i).isLiteral())
			sb.append("zo " + makeLojbanCmene(f.getSubject(i).toString()) + " ");
		else
			sb.append(synthesizeSubject(f.getSubject(i), resolver));
	}
	
	/**
	 * Add a fact to a string being built.
	 * @param f the fact to add (simple or composite)
	 * @param sb the string to build on
	 * @param resolver the resolver responsible
	 */
	protected void appendFact(IFact f, StringBuilder sb, IResolver resolver) {
		
		if(!f.isComposite()) {		// simple fact
			
			appendSubject(f, 0, sb, resolver);
			if(f.getSubject(0) != null)		// elide 'cu' if no x1 subject
				sb.append("cu ");
			sb.append(f.getRelation().toString().replace("\"", "") + " ");
			appendSubject(f, 1, sb, resolver);
			appendSubject(f, 2, sb, resolver);
			appendSubject(f, 3, sb, resolver);
			appendSubject(f, 4, sb, resolver);
			
		} else {					// complex fact
			
			// gek
			sb.append("g");
			switch(f.getConnective()) {
				case AND: sb.append("e"); break;
				case OR: sb.append("a"); break;
				case IFF: sb.append("o"); break;
				case WON: sb.append("u"); break;
			}
			if(f.getNegLeft())
				sb.append("nai");
			
			// first fact
			appendFact(f.getLeft(), sb, resolver);
			
			// gik
			sb.append("gi");
			if(f.getNegLeft())
				sb.append("nai");
			
			// second fact
			appendFact(f.getRight(), sb, resolver);
		}
	}
}