/*  Copyright 2005, Craig Andrews, Mike Itz, Martin Meyer
 
    This file is part of "Natural Language Interface Using First Order Logic"

    "Natural Language Interface Using First Order Logic" is free software;
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License, or (at your option) any later version.

    "Natural Language Interface Using First Order Logic" is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
package textToSpeech.voice;

import textToSpeech.exceptions.IllegalCharacterException;
import textToSpeech.exceptions.IllegalDiphthongException;

import com.sun.speech.freetts.en.us.CMULexicon;

public class LojbanLexicon extends CMULexicon {

	private String alphabet = "abcdefgijklmnoprstuvxyz.,'";

	public LojbanLexicon(String base) {
		super(base);
	}

	/*
	 * public boolean isSyllableBoundary(List syllablePhones, String[]
	 * wordPhones, int currentWordPhone) { // note that this code is just a hack
	 * as a test, it's not actually // accurate // if two consonants, it's a
	 * boundary
	 * 
	 * try { if (isConsonant(wordPhones[currentWordPhone]) &&
	 * isConsonant(wordPhones[currentWordPhone - 1])) return true; } catch
	 * (Exception e) { } // if vowel after vowel-consonant, it's a boundary try {
	 * if (!isConsonant(wordPhones[currentWordPhone]) &&
	 * isConsonant(wordPhones[currentWordPhone - 1]) &&
	 * !isConsonant(wordPhones[currentWordPhone - 1])) return true; } catch
	 * (Exception e) { } return false; }
	 */
	/**
	 * @see com.sun.speech.freetts.lexicon.Lexicon#getPhones(java.lang.String,
	 *      java.lang.String)
	 * 
	 * This implementation of getPhones(String, String) ignores any invalid
	 * letters (but prints an error). Illegal dipthongs result in an ah phone
	 * being returned.
	 */
	public String[] getPhones(String word, String partOfSpeech) {

		String[] result = new String[word.length()];
		for (int i = 0; i < word.length(); i++) {
			// we save the previous, current, and next characters just in case
			// they are needed.
			@SuppressWarnings("unused") char b = 0, c = 0, d = 0;
			if (i > 0)
				b = word.charAt(i - 1);
			c = word.charAt(i);
			if (!isLastChar(word, i)) {
				d = word.charAt(i + 1);
			} else {
				d = ',';
			}

			if (alphabet.indexOf(c) < 0) {
				throw new IllegalCharacterException(c);
			}

			if (!isConsonant(c) && !isConsonant(d)) {
				String[] phones = getDoubleVowelPhone(c, d).split(" ");
				for (int j = 0; j < phones.length; j++) {
					result[i] = phones[j];
					i++;
				}
				i--;
			} else if (c == 't' && d == 'c') {
				result[i] = "ch";
				i++;
			} else if  (c == 'd' && d == 'j') {
				result[i] = "jh";
				i++;
			} else {
				switch (c) {
				case '\'':
					// XXX: This case (') needs some thinking.
					// Although the ' is pronounced like an h (sorta), it
					// might
					// be necessary to change the phone of the previous
					// letter as well
					result[i] = "hh";
					break;
				case ',':
					// the comma only occurs between two vowels to prevent
					// them being interpreted as a dipthing. no phone is
					// necessary.
					result[i] = "pau";
					break;
				case '.':
					result[i] = "pau";
					break;

				// vowels first
				case 'a':
					// as in f_A_ther (specifically not h_A_t)
					result[i] = "ah";
					break;
				case 'e':
					// as in g_E_t
					result[i] = "eh";
					break;
				case 'i':
					// as in mach_I_ne or v_I_no (specifically not like
					// h_I_t)
					result[i] = "iy";
					// System.out.println("@@" + b + "@@" + c + "@@" + d +
					// "@@");
					break;
				case 'o':
					// as in b_O_ld or m_O_re (specifically not s_O_)
					result[i] = "ow";
					break;
				case 'u':
					// as in c_OO_l (specifically not b_U_t)
					result[i] = "uw";
					break;

				// special replacements
				case 'c':
					// sh in SH_ip
					result[i] = "sh";
					break;
				case 'g':
					// g as in G_um (specifically not G_em)
					result[i] = "g"; // this needs to be a hard G sound
					break;
				case 'j':
					// as in mea_S_ure or bon_J_ure
					result[i] = "zh"; // this phone might be wrong
					break;
				case 'x':
					// kh sound, like ba_CH, or KH_aled
					result[i] = "k";
					break;

				// regular consonents next
				case 'b':
					result[i] = "b";
					break;
				case 'd':
					result[i] = "d";
					break;
				case 'f':
					result[i] = "f";
					break;
				case 'k':
					result[i] = "k";
					break;
				case 'l':
					result[i] = "l";
					break;
				case 'm':
					result[i] = "m";
					break;
				case 'n':
					result[i] = "n";
					break;
				case 'p':
					result[i] = "p";
					break;
				case 'r':
					result[i] = "r";
					break;
				case 's':
					result[i] = "s";
					break;
				case 't':
					result[i] = "t";
					break;
				case 'v':
					result[i] = "v";
					break;
				case 'y':
					result[i] = "y";
					break;
				case 'z':
					result[i] = "z";
					break;
				}
			}
		}

		return result;
	}

	private boolean isLastChar(String word, int i) {
		return (i == word.length() - 1) ? true : false;

	}

	/**
	 * This method returns the phone for a dipthong. An invalid dipthong returns
	 * "pau", which is silent.
	 * 
	 * @param c
	 *            the first vowel
	 * @param d
	 *            the second vowel
	 * @return a phone which is part of the CMUDiphoneVoice
	 */
	private String getDoubleVowelPhone(char c, char d) {
		// TODO: fill in this method with actual intelligence
		switch (c) {
		case 'a':
			switch (d) {
			case 'i':
				return "ah iy"; // this needs to sound like "h_IGH"
			case 'u':
				return "ah ow"; // this needs to sound like "h_OW"
			}
		case 'e':
			switch (d) {
			case 'i':
				return "eh iy"; // this needs to sound like "h_EY"
			}
		case 'i':
			switch (d) {
			case 'a':
				return "iy aw"; // this needs to be like german "Ja", which
			// which sounds like "yaw"
			case 'e':
				return "iy aa"; // this needs to sound like "yeah"
			case 'u':
				return "iy uw"; // this needs to sound like "you"
			}
		case 'o':
			switch (d) {
			case 'i':
				return "ow iy"; // this needs to sound like "b_OY"
			}
		case 'u':
			switch (d) {
			case 'a':
				return "w aa"; // this needs to sound like "waah!"
			case 'e':
				return "w ey"; // this needs to sound like "q_UE_stion"
			case 'i':
				return "w iy"; // this needs to sound like "we"
			case 'o':
				return "w ow"; // this needs to sound like "q_UO_te
			}
		}

		throw new IllegalDiphthongException(c, d);
	}

	// / UTILITY ///

	/**
	 * Test to see if a character is a consonant
	 */
	public boolean isConsonant(char c) {
		boolean retval = true;
		if (c == 'a')
			retval = false;
		if (c == 'e')
			retval = false;
		if (c == 'i')
			retval = false;
		if (c == 'o')
			retval = false;
		if (c == 'u')
			retval = false;
		return retval;
	}

	// / SIMPLE ///

	public String[] getPhones(String word, String partOfSpeech, boolean useLTS) {
		// we're always using LTS (letter-to-sound rules) anyways
		return getPhones(word, partOfSpeech);
	}

	public boolean isLoaded() {
		return true;
	}
}
